"""Delete all character section data."""

from tqdm import tqdm

from django.core.management.base import BaseCommand

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.constants import IS_TESTING
from memberaudit.helpers import character_section_models
from memberaudit.models import Character
from memberaudit.tasks import update_all_characters

from . import get_input

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = str(__doc__)

    def add_arguments(self, parser):
        parser.add_argument(
            "--reload",
            action="store_true",
            help="Also start reloading",
        )
        parser.add_argument(
            "--all",
            action="store_true",
            help=(
                "Delete all character data, incl. from those "
                "which can not be reloaded from ESI"
            ),
        )
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_true",
            help="Do NOT prompt the user for input of any kind.",
        )

    def handle(self, *args, **options):
        if options["all"]:
            characters_query = Character.objects.all()
            orphan_text = "are included."
        else:
            characters_query = Character.objects.filter(
                is_disabled=False, eve_character__character_ownership__isnull=False
            )
            orphan_text = "are not included."

        character_count = characters_query.count()
        self.stdout.write(
            f"This will delete the section data for {character_count} characters. "
            f"Orphans and disabled characters {orphan_text}."
        )
        self.stdout.write(
            "All other local data incl. doctrines and the characters themselves will stay intact."
        )
        if options["reload"]:
            self.stdout.write("Will also start the task for reloading all characters.")

        self.stdout.write(
            self.style.NOTICE(
                "Please make sure your supervisors are shut down "
                "before running this command."
            )
        )
        if not options["noinput"]:
            user_input = get_input("Are you sure you want to proceed? (y/N)?")
        else:
            user_input = "y"

        if user_input.lower() != "y":
            self.stdout.write(self.style.WARNING("Aborted"))
            return

        logger.info(
            "Running command reset_characters for %s characters.", character_count
        )
        section_models = character_section_models()
        for character in tqdm(
            characters_query.iterator(),
            desc="Deleting characters",
            total=character_count,
            disable=IS_TESTING,
        ):
            for model_class in section_models:
                if hasattr(model_class, "character"):
                    model_class.objects.filter(character=character).delete()

        if options["reload"]:
            update_all_characters.delay(force_update=True, ignore_stale=True)
            self.stdout.write("Started task to reload all character data")

        self.stdout.write(self.style.SUCCESS("Done"))



from io import StringIO
from unittest.mock import patch

from django.core.management import call_command

from app_utils.testing import NoSocketsTestCase

from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.utils import create_memberaudit_character

MODULE_PATH = "memberaudit.management.commands.memberaudit_update_characters"


@patch(MODULE_PATH + ".tasks")
class TestUpdateCharacters(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_should_reset_section_data_for_characters(self, mock_tasks):
        # given
        create_memberaudit_character(1001)
        out = StringIO()

        # when
        call_command("memberaudit_update_characters", "--noinput", "assets", stdout=out)

        # then
        self.assertTrue(mock_tasks.update_character_assets.apply_async.called)



from app_utils.testing import NoSocketsTestCase, create_authgroup

from .testdata.factories import create_compliance_group_designation
from .testdata.load_entities import load_entities


class TestSignals(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_should_prevent_making_compliance_group_non_internal(self):
        # given
        group = create_authgroup()
        create_compliance_group_designation(group)
        # when
        group.authgroup.internal = False
        group.authgroup.save()
        # then
        group.refresh_from_db()
        self.assertTrue(group.authgroup.internal)

    def test_should_allow_making_other_groups_non_internal(self):
        # given
        group = create_authgroup()
        # when
        group.authgroup.internal = False
        group.authgroup.save()
        # then
        group.refresh_from_db()
        self.assertFalse(group.authgroup.internal)

    def test_should_allow_creating_non_internal_groups(self):
        # when
        group = create_authgroup(internal=False)
        # then
        group.refresh_from_db()
        self.assertFalse(group.authgroup.internal)



"""Deleting all unresolved and orphaned EveEntities.

This tool was written to remove invalid EveEntities created by a bug (see issue #156).
"""

from tqdm import tqdm

from django.contrib.admin.utils import NestedObjects
from django.core.management.base import BaseCommand, CommandParser
from django.db import DatabaseError
from django.db.models import QuerySet
from eveuniverse.models import EveEntity

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.constants import IS_TESTING

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Command(BaseCommand):
    help = str(__doc__)

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--no-input", "--noinput", action="store_true", help="Skip all user input"
        )

    def handle(self, *args, **options):
        objs_to_delete = _gather_objs_to_delete()
        to_delete_count = objs_to_delete.count()

        if not to_delete_count:
            self.stdout.write(
                self.style.SUCCESS(
                    "No matching EveEntity objects found. No fix required."
                )
            )
            return

        response = self._ask_user(options, to_delete_count)
        if response.lower() == "n":
            self.stdout.write(self.style.WARNING("Aborted"))
            return

        deleted_count, error_count = _delete_objects(objs_to_delete)

        if error_count:
            self.stdout.write(
                self.style.WARNING(
                    f"{error_count} EveEntity objects could not be deleted. "
                    "See extension log for details."
                )
            )

        self.stdout.write(
            self.style.SUCCESS(f"Deleted {deleted_count} EveEntity objects.")
        )

    def _ask_user(self, options, to_delete_count) -> str:
        if options["no_input"]:
            return "y"

        self.stdout.write(
            f"Found {to_delete_count} unresolved and orphaned EveEntity objects, "
            "which can be safely removed."
        )
        self.stdout.write("Do you want to delete these objects? ", ending="")
        response = input("(Y/n)")
        return response


def _gather_objs_to_delete() -> QuerySet:
    objs = EveEntity.objects.filter(name="", category=None)
    if not objs.exists():
        return EveEntity.objects.none()

    to_delete_ids = []
    objs_with_progress_bar = tqdm(
        objs,
        desc="Checking relations",
        total=objs.count(),
        disable=IS_TESTING,
        leave=False,
    )
    for obj in objs_with_progress_bar:
        collector = NestedObjects(using="default")
        collector.collect([obj])
        related_model_count = collector.data.keys()
        if len(related_model_count) == 1:
            to_delete_ids.append(obj.id)

    return EveEntity.objects.filter(id__in=to_delete_ids)


def _delete_objects(objs_to_delete: QuerySet[EveEntity]) -> int:
    error_count = 0
    deleted_count = 0
    objs_with_progress_bar = tqdm(
        objs_to_delete,
        desc="Deleting objects",
        total=objs_to_delete.count(),
        leave=False,
        disable=IS_TESTING,
    )
    for obj in objs_with_progress_bar:
        try:
            obj.delete()
        except DatabaseError:
            logger.exception("Failed to delete EveEntity object with ID %d", obj.id)
            error_count += 1
        else:
            deleted_count += 1

    logger.info("Deleted %d EveEntity objects.", deleted_count)
    return deleted_count, error_count



import math
from dataclasses import dataclass, field
from typing import List, Set

from tqdm import tqdm

from django.core.management.base import BaseCommand
from django.db import DatabaseError

from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from memberaudit import __title__, tasks
from memberaudit.constants import IS_TESTING
from memberaudit.models import (
    Character,
    CharacterAsset,
    CharacterContract,
    CharacterJumpClone,
    CharacterLocation,
    CharacterUpdateStatus,
    CharacterWalletTransaction,
    Location,
)

from . import get_input

DEFAULT_BATCH_SIZE_FETCH = 10_000
DEFAULT_BATCH_SIZE_UPDATE = 100

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@dataclass
class CharacterPkContainer:
    """Container for character PKs."""

    assets: Set[int] = field(default_factory=set)
    clones: Set[int] = field(default_factory=set)
    contracts: Set[int] = field(default_factory=set)
    locations: Set[int] = field(default_factory=set)
    transactions: Set[int] = field(default_factory=set)

    def __bool__(self) -> bool:
        return (
            bool(self.assets)
            or bool(self.clones)
            or bool(self.contracts)
            or bool(self.locations)
            or bool(self.transactions)
        )

    def __len__(self) -> int:
        return len(self.all())

    def all(self) -> Set[int]:
        """Return all character PKs."""
        return (
            self.assets
            | self.clones
            | self.contracts
            | self.locations
            | self.transactions
        )


class Command(BaseCommand):
    help = "Remove invalid locations and corrupted data caused by issue #153"

    def add_arguments(self, parser):
        parser.add_argument(
            "--batch-size-fetch",
            type=int,
            default=DEFAULT_BATCH_SIZE_FETCH,
            help="Maximum number of invalid locations fetched per batch",
        )

        parser.add_argument(
            "--batch-size-update",
            type=int,
            default=DEFAULT_BATCH_SIZE_UPDATE,
            help="Maximum number of invalid locations fixed per batch",
        )

        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_true",
            help="Do NOT prompt the user for input of any kind",
        )

        parser.add_argument(
            "--verbose-log",
            action="store_true",
            help="Log additional details",
        )

        parser.add_argument(
            "--exclude-locations",
            nargs="+",
            type=int,
            help="Exclude locations given by their IDs",
        )

    def handle(self, *args, **options):
        logger.info("Started command for fixing invalid locations")

        batch_size = options["batch_size_fetch"]
        exclude_location_ids = (
            set(options["exclude_locations"]) if options["exclude_locations"] else set()
        )
        is_verbose_log = options["verbose_log"]

        invalid_location_ids = find_invalid_locations(
            batch_size=batch_size,
            exclude_location_ids=exclude_location_ids,
            is_verbose_log=is_verbose_log,
        )

        if not invalid_location_ids:
            self.stdout.write(self.style.SUCCESS("No invalid locations found"))
            return

        if not options["noinput"]:
            self.stdout.write(
                f"This command will remove {len(invalid_location_ids):,} "
                "invalid locations "
                "and fix related character data corruption caused by issue #153. "
                "Details will be logged to the extensions log"
            )
            self.stdout.write("This process can take a while to complete")
            user_input = get_input("Are you sure you want to proceed (Y/n)?")
        else:
            user_input = "y"

        if user_input.lower() == "n":
            self.stdout.write(self.style.WARNING("Aborted"))
            return

        self.stdout.write("")

        character_pks_all = self._fix_data_corruption_and_remove_invalid_locations(
            invalid_location_ids, options
        )

        remaining_invalid_location_ids = find_invalid_locations(
            batch_size=batch_size,
            exclude_location_ids=exclude_location_ids,
            is_verbose_log=is_verbose_log,
        )
        self.stdout.write(
            f"{len(remaining_invalid_location_ids)} invalid locations remaining"
        )

        characters_updateable_pks = identify_updateable_characters(
            character_pks=character_pks_all, is_verbose_log=is_verbose_log
        )

        if not characters_updateable_pks:
            self.stdout.write(self.style.SUCCESS("Done"))
            return

        self.stdout.write(
            f"Data for up to {len(characters_updateable_pks):,} characters may "
            "have been disrupted by invalid locations"
        )
        if not options["noinput"]:
            self.stdout.write(
                "Do you want to (s)tart an immediate update for these characters?"
            )
            self.stdout.write(
                "Or do you want to (w)ait for the update to happen "
                "with the regular schedule?"
            )
            user_input = get_input("(s/W)?")
        else:
            user_input = "s"

        if user_input.lower() == "s":
            start_character_updates(characters_updateable_pks)
            msg = (
                "Immediate updates has been started for "
                f"{len(characters_updateable_pks):,} characters"
            )
            logger.info(msg)
            self.stdout.write(msg)

        else:
            self.stdout.write("Characters will be updated with the next regular update")

        self.stdout.write(self.style.SUCCESS("Done"))

    def _fix_data_corruption_and_remove_invalid_locations(
        self, invalid_location_ids: List[int], options: dict
    ) -> CharacterPkContainer:
        invalid_location_count = len(invalid_location_ids)
        batch_size: int = options["batch_size_update"]
        logger.info(
            "Started fixing %d invalid locations (batch size: %d)",
            invalid_location_count,
            batch_size,
        )

        unknown_location, _ = Location.objects.get_or_create_unknown_location()  # type: ignore
        character_pks = CharacterPkContainer()
        batch_count = math.ceil(invalid_location_count / batch_size)
        removed_location_ids = []
        error_count = 0

        for location_ids_chunk in tqdm(
            chunks(invalid_location_ids, size=batch_size),
            desc="Fixing locations",
            total=batch_count,
            leave=False,
            unit_scale=batch_size,
            disable=IS_TESTING,
        ):
            is_success = fix_invalid_locations(
                location_ids=location_ids_chunk,
                character_pks=character_pks,
                unknown_location=unknown_location,
                is_verbose_log=options["verbose_log"],
            )
            if is_success:
                removed_location_ids += location_ids_chunk
            else:
                error_count += 1

        msg = (
            f"Process completed: Removed {len(removed_location_ids):,} "
            "invalid locations and repaired related character data"
        )
        logger.info(msg)
        self.stdout.write(msg)

        if error_count:
            msg = f"{error_count:,} errors occurred. Please check logs for details"
            self.stdout.write(self.style.ERROR(msg))
            logger.warning(msg)

        self.stdout.write("")
        return character_pks


def find_invalid_locations(
    batch_size: int, exclude_location_ids: Set[int], is_verbose_log: bool
) -> List[int]:
    """Return IDs of invalid locations.
    An empty list means no invalid locations where found.
    """
    asset_item_ids = set(CharacterAsset.objects.values_list("item_id", flat=True))
    if exclude_location_ids:
        asset_item_ids -= set(exclude_location_ids)
        logger.info("Ignoring location IDs: %s", exclude_location_ids)

    asset_items_count = len(asset_item_ids)
    logger.info(
        "Looking for invalid locations among %d asset items (batch size: %d)",
        asset_items_count,
        batch_size,
    )
    batch_count = math.ceil(asset_items_count / batch_size)
    invalid_location_ids = []
    for asset_item_ids_chunk in tqdm(
        chunks(list(asset_item_ids), size=batch_size),
        desc="Finding invalid locations",
        total=batch_count,
        leave=False,
        unit_scale=batch_size,
        disable=IS_TESTING,
    ):
        invalid_location_ids += _find_invalid_locations_chunk(asset_item_ids_chunk)

    invalid_location_ids.sort()

    if is_verbose_log:
        logger.info(
            "Found %d invalid locations: %s",
            len(invalid_location_ids),
            invalid_location_ids,
        )
    else:
        logger.info("Found %d invalid locations", len(invalid_location_ids))

    return invalid_location_ids


def _find_invalid_locations_chunk(asset_item_ids_chunk: List[int]) -> List[int]:
    invalid_locations = Location.objects.filter(id__in=asset_item_ids_chunk)
    invalid_location_ids = list(invalid_locations.values_list("id", flat=True))
    return invalid_location_ids


def fix_invalid_locations(
    location_ids: List[int],
    character_pks: CharacterPkContainer,
    unknown_location: Location,
    is_verbose_log: bool,
) -> bool:
    """Fix corruption in character data and remove invalid locations.

    Return True, when successful, else False.
    """
    try:
        character_pks.assets |= _fix_corrupted_character_section(
            location_ids=location_ids,
            unknown_location=unknown_location,
            section=Character.UpdateSection.ASSETS,
            model_class=CharacterAsset,
            is_verbose_log=is_verbose_log,
        )
        character_pks.clones |= _fix_corrupted_character_section(
            location_ids=location_ids,
            unknown_location=unknown_location,
            section=Character.UpdateSection.JUMP_CLONES,
            model_class=CharacterJumpClone,
            is_verbose_log=is_verbose_log,
        )
        character_pks.contracts |= _fix_corrupted_character_section(
            location_ids=location_ids,
            unknown_location=unknown_location,
            section=Character.UpdateSection.CONTRACTS,
            model_class=CharacterContract,
            field_name="start_location",
            is_verbose_log=is_verbose_log,
        )
        character_pks.contracts |= _fix_corrupted_character_section(
            location_ids=location_ids,
            unknown_location=unknown_location,
            section=Character.UpdateSection.CONTRACTS,
            model_class=CharacterContract,
            field_name="end_location",
            is_verbose_log=is_verbose_log,
        )
        character_pks.locations |= _fix_corrupted_character_section(
            location_ids=location_ids,
            unknown_location=unknown_location,
            section=Character.UpdateSection.LOCATION,
            model_class=CharacterLocation,
            is_verbose_log=is_verbose_log,
        )
        character_pks.transactions |= _fix_corrupted_character_section(
            location_ids=location_ids,
            unknown_location=unknown_location,
            section=Character.UpdateSection.WALLET_TRANSACTIONS,
            model_class=CharacterWalletTransaction,
            is_verbose_log=is_verbose_log,
        )
        locations_chunk = Location.objects.filter(id__in=location_ids)
        locations_chunk._raw_delete(locations_chunk.db)  # type: ignore

    except DatabaseError:
        logger.exception(
            "Failed to remove %d invalid locations. Skipping to next chunk",
            len(location_ids),
        )
        return False

    logger.info("Deleted %d invalid locations", len(location_ids))
    return True


def _fix_corrupted_character_section(
    section: Character.UpdateSection,
    location_ids: List[int],
    unknown_location: Location,
    model_class: type,
    field_name: str = "location",
    is_verbose_log: bool = False,
) -> Set[int]:
    """Entangle locations from a character section of any related characters.

    Return character PKs, when updates where successful, else empty set.
    """
    params_filter = {f"{field_name}__in": location_ids}
    corrupted_objs = model_class.objects.filter(**params_filter)  # type: ignore
    corrupted_objs_count = corrupted_objs.count()
    if not corrupted_objs_count:
        return set()

    character_pks = list(
        sorted(corrupted_objs.values_list("character__pk", flat=True).distinct())
    )
    params_update = {field_name: unknown_location}
    corrupted_objs.update(**params_update)

    CharacterUpdateStatus.objects.filter(
        character__pk__in=character_pks, section=section
    ).update(content_hash_1="", content_hash_2="", content_hash_3="")
    if is_verbose_log:
        logger.info(
            "Removed %d invalid locations from %d corrupted %s across %d characters: %s",
            len(location_ids),
            corrupted_objs_count,
            section.label,
            len(character_pks),
            character_pks,
        )
    else:
        logger.info(
            "Removed %d invalid locations from %d corrupted %s across %d characters",
            len(location_ids),
            corrupted_objs_count,
            section.label,
            len(character_pks),
        )

    return set(character_pks)


def identify_updateable_characters(
    character_pks: CharacterPkContainer, is_verbose_log: bool
) -> CharacterPkContainer:
    """Return selection of character PKs, which can be updated."""
    params = {}
    for section in ["assets", "clones", "contracts", "locations", "transactions"]:
        params[section] = set(
            Character.objects.filter(
                pk__in=getattr(character_pks, section),
                is_disabled=False,
                eve_character__character_ownership__isnull=False,
            ).values_list("pk", flat=True)
        )

    updateable_character_pks = CharacterPkContainer(**params)

    if is_verbose_log:
        logger.info(
            "From %d repaired characters, %d can be updated: %s",
            len(character_pks),
            len(updateable_character_pks),
            updateable_character_pks,
        )
    else:
        logger.info(
            "From %d repaired characters, %d can be updated",
            len(character_pks),
            len(updateable_character_pks),
        )

    return updateable_character_pks


def start_character_updates(character_pks: CharacterPkContainer) -> None:
    """Start character section updates for characters as needed."""
    for character_pk in character_pks.assets:
        tasks.update_character_assets.apply_async(
            kwargs={"character_pk": character_pk, "force_update": True},
            priority=tasks.MEMBERAUDIT_TASKS_LOW_PRIORITY,
        )  # type: ignore

    for character_pk in character_pks.contracts:
        tasks.update_character_contracts.apply_async(
            kwargs={"character_pk": character_pk, "force_update": True},
            priority=tasks.MEMBERAUDIT_TASKS_LOW_PRIORITY,
        )  # type: ignore

    for character_pk in character_pks.clones:
        tasks.update_character_jump_clones.apply_async(
            kwargs={"character_pk": character_pk, "force_update": True},
            priority=tasks.MEMBERAUDIT_TASKS_LOW_PRIORITY,
        )  # type: ignore

    for character_pk in character_pks.locations:
        tasks.update_character_location.apply_async(
            kwargs={"character_pk": character_pk, "force_update": True},
            priority=tasks.MEMBERAUDIT_TASKS_LOW_PRIORITY,
        )  # type: ignore

    for character_pk in character_pks.transactions:
        tasks.update_character_wallet_transactions.apply_async(
            kwargs={"character_pk": character_pk, "force_update": True},
            priority=tasks.MEMBERAUDIT_TASKS_LOW_PRIORITY,
        )  # type: ignore



from io import StringIO
from unittest.mock import patch

from django.core.management import call_command
from django.db import IntegrityError
from django.test import TestCase

from allianceauth.eveonline.models import EveCharacter

from memberaudit.management.commands import memberaudit_fix_locations
from memberaudit.models import Character, Location
from memberaudit.tests.testdata.factories import (
    create_character,
    create_character_asset,
    create_character_contract_courier,
    create_character_jump_clone,
    create_character_location,
    create_character_update_status,
    create_character_wallet_transaction,
    create_location,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.utils import create_memberaudit_character

MODULE_PATH = "memberaudit.management.commands.memberaudit_fix_locations"


@patch(MODULE_PATH + ".tasks.update_character_wallet_transactions", spec=True)
@patch(MODULE_PATH + ".tasks.update_character_location", spec=True)
@patch(MODULE_PATH + ".tasks.update_character_jump_clones", spec=True)
@patch(MODULE_PATH + ".tasks.update_character_contracts", spec=True)
@patch(MODULE_PATH + ".tasks.update_character_assets", spec=True)
class TestFixInvalidLocations(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    def _extract_character_from_called_task(self, mock_task):
        assets_task_calls = [
            o[1]["kwargs"] for o in mock_task.apply_async.call_args_list
        ]
        self.assertEqual(len(assets_task_calls), 1)
        params = assets_task_calls[0]
        return params["character_pk"]

    def test_should_do_nothing_when_no_invalid_locations(self, *args, **kwargs):
        # given
        asset = create_character_asset(
            character=self.character_1001, location=create_location()
        )
        contract = create_character_contract_courier(
            character=self.character_1001,
            start_location=create_location(),
            end_location=create_location(),
        )
        location = create_character_location(
            character=self.character_1001, location=create_location()
        )
        jump_clone = create_character_jump_clone(
            character=self.character_1001, location=create_location()
        )
        wallet = create_character_wallet_transaction(
            character=self.character_1001, location=create_location()
        )
        out = StringIO()

        # when
        call_command("memberaudit_fix_locations", "--noinput", stdout=out)

        # then
        asset.refresh_from_db()
        self.assertTrue(asset.location)
        contract.refresh_from_db()
        self.assertTrue(contract.start_location)
        self.assertTrue(contract.end_location)
        location.refresh_from_db()
        self.assertTrue(location.location)
        jump_clone.refresh_from_db()
        self.assertTrue(jump_clone.location)
        wallet.refresh_from_db()
        self.assertTrue(wallet.location)

    def test_should_delete_invalid_locations_and_fix_related_assets(
        self,
        mock_task_update_character_assets,
        mock_task_update_character_contracts,
        mock_task_update_character_jump_clones,
        mock_task_update_character_location,
        mock_task_update_character_wallet_transactions,
    ):
        # given characters
        character_1002 = create_memberaudit_character(1002)
        character_1002.is_disabled = True
        character_1002.save()
        character_1101 = create_character(  # orphan
            EveCharacter.objects.get(character_id=1101)
        )
        character_1003 = create_memberaudit_character(1003)  # no corruption

        # given locations
        valid_location_1 = create_location()
        valid_location_2 = create_location()
        invalid_location_1 = create_location()
        invalid_location_2 = create_location()

        # given assets
        normal_asset_1001 = create_character_asset(
            item_id=invalid_location_1.id,
            character=self.character_1001,
            location=valid_location_1,
        )
        corrupted_asset_1001 = create_character_asset(
            character=self.character_1001, location=invalid_location_1
        )
        status_assets_1001 = create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.ASSETS,
            content_hash_1="some_data",
        )

        corrupted_asset_1002 = create_character_asset(
            character=character_1002, location=invalid_location_1
        )
        status_assets_1002 = create_character_update_status(
            character=character_1002,
            section=Character.UpdateSection.ASSETS,
            content_hash_1="some_data",
        )

        corrupted_asset_1101 = create_character_asset(
            character=character_1101, location=invalid_location_1
        )
        status_assets_1101 = create_character_update_status(
            character=character_1101,
            section=Character.UpdateSection.ASSETS,
            content_hash_1="some_data",
        )
        corrupted_asset_1101 = create_character_jump_clone(
            character=character_1101, location=invalid_location_1
        )

        normal_asset_1003 = create_character_asset(
            item_id=invalid_location_2.id,
            character=character_1003,
            location=valid_location_1,
        )

        # given clones
        normal_clone_1001 = create_character_jump_clone(
            character=self.character_1001, location=valid_location_1
        )
        corrupted_clone_1001 = create_character_jump_clone(
            character=self.character_1001, location=invalid_location_1
        )
        status_clones_1001 = create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.JUMP_CLONES,
            content_hash_1="some_data",
        )

        corrupted_clone_1002 = create_character_jump_clone(
            character=character_1002, location=invalid_location_1
        )
        status_clones_1002 = create_character_update_status(
            character=character_1002,
            section=Character.UpdateSection.JUMP_CLONES,
            content_hash_1="some_data",
        )

        corrupted_clone_1101 = create_character_jump_clone(
            character=character_1101, location=invalid_location_1
        )
        status_clones_1101 = create_character_update_status(
            character=character_1101,
            section=Character.UpdateSection.JUMP_CLONES,
            content_hash_1="some_data",
        )

        # given character locations
        corrupted_location_1001 = create_character_location(
            character=self.character_1001, location=invalid_location_1
        )
        status_location_1001 = create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.LOCATION,
            content_hash_1="some_data",
        )

        corrupted_location_1002 = create_character_location(
            character=character_1002, location=invalid_location_1
        )
        status_location_1002 = create_character_update_status(
            character=character_1002,
            section=Character.UpdateSection.LOCATION,
            content_hash_1="some_data",
        )

        normal_location_1003 = create_character_location(
            character=character_1003, location=valid_location_1
        )
        status_location_1003 = create_character_update_status(
            character=character_1003,
            section=Character.UpdateSection.LOCATION,
            content_hash_1="some_data",
        )

        corrupted_location_1101 = create_character_location(
            character=character_1101, location=invalid_location_1
        )
        status_location_1101 = create_character_update_status(
            character=character_1101,
            section=Character.UpdateSection.LOCATION,
            content_hash_1="some_data",
        )

        # given wallet transactions
        normal_transaction_1001 = create_character_wallet_transaction(
            character=self.character_1001, location=valid_location_1
        )
        corrupted_transaction_1001 = create_character_wallet_transaction(
            character=self.character_1001, location=invalid_location_1
        )
        status_transactions_1001 = create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.WALLET_TRANSACTIONS,
            content_hash_1="some_data",
        )

        corrupted_transaction_1002 = create_character_wallet_transaction(
            character=character_1002, location=invalid_location_1
        )
        status_transactions_1002 = create_character_update_status(
            character=character_1002,
            section=Character.UpdateSection.WALLET_TRANSACTIONS,
            content_hash_1="some_data",
        )

        corrupted_transaction_1101 = create_character_wallet_transaction(
            character=character_1101, location=invalid_location_1
        )
        status_transactions_1101 = create_character_update_status(
            character=character_1101,
            section=Character.UpdateSection.WALLET_TRANSACTIONS,
            content_hash_1="some_data",
        )

        # given courier contracts
        normal_contract_1001 = create_character_contract_courier(
            character=self.character_1001,
            start_location=valid_location_1,
            end_location=valid_location_2,
        )
        corrupted_contract_1001 = create_character_contract_courier(
            character=self.character_1001,
            start_location=invalid_location_1,
            end_location=invalid_location_2,
        )
        status_contracts_1001 = create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.CONTRACTS,
            content_hash_1="some_data",
        )

        corrupted_contract_1002 = create_character_contract_courier(
            character=character_1002,
            start_location=invalid_location_1,
            end_location=invalid_location_2,
        )
        status_contracts_1002 = create_character_update_status(
            character=character_1002,
            section=Character.UpdateSection.CONTRACTS,
            content_hash_1="some_data",
        )

        corrupted_contract_1101 = create_character_contract_courier(
            character=character_1101,
            start_location=invalid_location_1,
            end_location=invalid_location_2,
        )
        status_contracts_1101 = create_character_update_status(
            character=character_1101,
            section=Character.UpdateSection.CONTRACTS,
            content_hash_1="some_data",
        )
        out = StringIO()

        # when
        call_command("memberaudit_fix_locations", "--noinput", stdout=out)

        # then locations
        location_ids = set(Location.objects.values_list("id", flat=True))
        self.assertSetEqual(
            location_ids,
            {valid_location_1.id, valid_location_2.id, Location.LOCATION_UNKNOWN_ID},
        )

        # then assets
        normal_asset_1001.refresh_from_db()
        self.assertEqual(normal_asset_1001.location, valid_location_1)
        corrupted_asset_1001.refresh_from_db()
        self.assertEqual(corrupted_asset_1001.location.id, Location.LOCATION_UNKNOWN_ID)
        status_assets_1001.refresh_from_db()
        self.assertFalse(status_assets_1001.content_hash_1)

        corrupted_asset_1002.refresh_from_db()
        self.assertEqual(corrupted_asset_1002.location.id, Location.LOCATION_UNKNOWN_ID)
        status_assets_1002.refresh_from_db()
        self.assertFalse(status_assets_1002.content_hash_1)

        corrupted_asset_1101.refresh_from_db()
        self.assertEqual(corrupted_asset_1101.location.id, Location.LOCATION_UNKNOWN_ID)
        status_assets_1101.refresh_from_db()
        self.assertFalse(status_assets_1101.content_hash_1)

        normal_asset_1003.refresh_from_db()
        self.assertEqual(normal_asset_1003.location, valid_location_1)

        called_character_pk_in_task = self._extract_character_from_called_task(
            mock_task_update_character_assets
        )
        self.assertEqual(called_character_pk_in_task, self.character_1001.pk)

        # then clones
        normal_clone_1001.refresh_from_db()
        self.assertEqual(normal_clone_1001.location, valid_location_1)
        corrupted_clone_1001.refresh_from_db()
        self.assertEqual(corrupted_clone_1001.location.id, Location.LOCATION_UNKNOWN_ID)
        status_clones_1001.refresh_from_db()
        self.assertFalse(status_clones_1001.content_hash_1)

        corrupted_clone_1002.refresh_from_db()
        self.assertEqual(corrupted_clone_1002.location.id, Location.LOCATION_UNKNOWN_ID)
        status_clones_1002.refresh_from_db()
        self.assertFalse(status_clones_1002.content_hash_1)

        corrupted_clone_1101.refresh_from_db()
        self.assertEqual(corrupted_clone_1101.location.id, Location.LOCATION_UNKNOWN_ID)
        status_clones_1101.refresh_from_db()
        self.assertFalse(status_clones_1101.content_hash_1)

        called_character_pk_in_task = self._extract_character_from_called_task(
            mock_task_update_character_jump_clones
        )
        self.assertEqual(called_character_pk_in_task, self.character_1001.pk)

        # then character locations
        corrupted_location_1001.refresh_from_db()
        self.assertEqual(
            corrupted_location_1001.location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_location_1001.refresh_from_db()
        self.assertFalse(status_location_1001.content_hash_1)

        corrupted_location_1002.refresh_from_db()
        self.assertEqual(
            corrupted_location_1002.location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_location_1002.refresh_from_db()
        self.assertFalse(status_location_1002.content_hash_1)

        normal_location_1003.refresh_from_db()
        self.assertEqual(normal_location_1003.location, valid_location_1)
        self.assertTrue(status_location_1003.content_hash_1)

        corrupted_location_1101.refresh_from_db()
        self.assertEqual(
            corrupted_location_1101.location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_location_1101.refresh_from_db()
        self.assertFalse(status_location_1101.content_hash_1)

        called_character_pk_in_task = self._extract_character_from_called_task(
            mock_task_update_character_location
        )
        self.assertEqual(called_character_pk_in_task, self.character_1001.pk)

        # then wallet transactions
        normal_transaction_1001.refresh_from_db()
        self.assertEqual(normal_transaction_1001.location, valid_location_1)
        corrupted_transaction_1001.refresh_from_db()
        self.assertEqual(
            corrupted_transaction_1001.location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_transactions_1001.refresh_from_db()
        self.assertFalse(status_transactions_1001.content_hash_1)

        corrupted_transaction_1002.refresh_from_db()
        self.assertEqual(
            corrupted_transaction_1002.location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_transactions_1002.refresh_from_db()
        self.assertFalse(status_transactions_1002.content_hash_1)

        corrupted_transaction_1101.refresh_from_db()
        self.assertEqual(
            corrupted_transaction_1101.location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_transactions_1101.refresh_from_db()
        self.assertFalse(status_transactions_1101.content_hash_1)

        called_character_pk_in_task = self._extract_character_from_called_task(
            mock_task_update_character_wallet_transactions
        )
        self.assertEqual(called_character_pk_in_task, self.character_1001.pk)

        # then contracts
        normal_contract_1001.refresh_from_db()
        self.assertEqual(normal_contract_1001.start_location, valid_location_1)
        self.assertEqual(normal_contract_1001.end_location, valid_location_2)
        corrupted_contract_1001.refresh_from_db()
        self.assertEqual(
            corrupted_contract_1001.start_location.id, Location.LOCATION_UNKNOWN_ID
        )
        self.assertEqual(
            corrupted_contract_1001.end_location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_contracts_1001.refresh_from_db()
        self.assertFalse(status_contracts_1001.content_hash_1)

        corrupted_contract_1002.refresh_from_db()
        self.assertEqual(
            corrupted_contract_1002.start_location.id, Location.LOCATION_UNKNOWN_ID
        )
        self.assertEqual(
            corrupted_contract_1002.end_location.id, Location.LOCATION_UNKNOWN_ID
        )
        status_contracts_1002.refresh_from_db()
        self.assertFalse(status_contracts_1002.content_hash_1)

        corrupted_contract_1101.refresh_from_db()
        self.assertEqual(
            corrupted_contract_1101.start_location.id, Location.LOCATION_UNKNOWN_ID
        )
        self.assertEqual(
            corrupted_contract_1101.end_location.id, Location.LOCATION_UNKNOWN_ID
        )

        status_contracts_1101.refresh_from_db()
        self.assertFalse(status_contracts_1101.content_hash_1)

        called_character_pk_in_task = self._extract_character_from_called_task(
            mock_task_update_character_contracts
        )
        self.assertEqual(called_character_pk_in_task, self.character_1001.pk)

    def test_should_ignore_db_issue_when_fixing_section(self, *args, **kwargs):
        # given
        valid_location = create_location()
        invalid_location = create_location()

        asset = create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location.id,
        )
        corrupted_asset = create_character_asset(
            character=self.character_1001, location=invalid_location
        )
        out = StringIO()

        # when
        with patch(
            MODULE_PATH + ".CharacterAsset",
            wraps=memberaudit_fix_locations.CharacterAsset,
        ) as mock:
            mock.objects.filter.return_value.update.side_effect = IntegrityError
            call_command("memberaudit_fix_locations", "--noinput", stdout=out)

        # then
        asset.refresh_from_db()
        self.assertTrue(asset.location)
        corrupted_asset.refresh_from_db()
        self.assertEqual(corrupted_asset.location.id, invalid_location.id)

    def test_should_start_update_tasks_for_needed_characters_only(
        self,
        mock_task_update_character_assets,
        mock_task_update_character_contracts,
        mock_task_update_character_jump_clones,
        mock_task_update_character_location,
        mock_task_update_character_wallet_transactions,
    ):
        # given
        character_pks = memberaudit_fix_locations.CharacterPkContainer()
        assets_pk = 1001
        character_pks.assets.add(assets_pk)
        clones_pk = 1002
        character_pks.clones.add(clones_pk)
        contracts_pk = 1002
        character_pks.contracts.add(contracts_pk)
        location_pk = 1003
        character_pks.locations.add(location_pk)
        transactions_pk = 1004
        character_pks.transactions.add(transactions_pk)

        # when
        memberaudit_fix_locations.start_character_updates(character_pks)

        # then
        character_pk_called = self._extract_character_from_called_task(
            mock_task_update_character_assets
        )
        self.assertEqual(character_pk_called, assets_pk)

        character_pk_called = self._extract_character_from_called_task(
            mock_task_update_character_contracts
        )
        self.assertEqual(character_pk_called, contracts_pk)

        character_pk_called = self._extract_character_from_called_task(
            mock_task_update_character_jump_clones
        )
        self.assertEqual(character_pk_called, clones_pk)

        character_pk_called = self._extract_character_from_called_task(
            mock_task_update_character_location
        )
        self.assertEqual(character_pk_called, location_pk)

        character_pk_called = self._extract_character_from_called_task(
            mock_task_update_character_wallet_transactions
        )
        self.assertEqual(character_pk_called, transactions_pk)

    def test_should_allow_excluding_locations(self, *args, **kwargs):
        # given
        valid_location = create_location()
        invalid_location_1 = create_location()
        invalid_location_2 = create_location()
        invalid_location_3 = create_location()

        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_1.id,
        )
        corrupted_asset_1 = create_character_asset(
            character=self.character_1001, location=invalid_location_1
        )
        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_2.id,
        )
        corrupted_asset_2 = create_character_asset(
            character=self.character_1001, location=invalid_location_2
        )
        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_3.id,
        )
        corrupted_asset_3 = create_character_asset(
            character=self.character_1001, location=invalid_location_3
        )

        out = StringIO()

        # when
        call_command(
            "memberaudit_fix_locations",
            "--noinput",
            "--exclude-locations",
            str(invalid_location_1.id),
            str(invalid_location_3.id),
            stdout=out,
        )

        # then
        corrupted_asset_1.refresh_from_db()
        self.assertEqual(corrupted_asset_1.location.id, invalid_location_1.id)

        corrupted_asset_2.refresh_from_db()
        self.assertEqual(corrupted_asset_2.location.id, Location.LOCATION_UNKNOWN_ID)

        corrupted_asset_3.refresh_from_db()
        self.assertEqual(corrupted_asset_3.location.id, invalid_location_3.id)

    def test_should_fix_locations_in_batches(self, *args, **kwargs):
        # given
        valid_location = create_location()
        invalid_location_1 = create_location()
        invalid_location_2 = create_location()
        invalid_location_3 = create_location()

        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_1.id,
        )
        corrupted_asset_1 = create_character_asset(
            character=self.character_1001, location=invalid_location_1
        )
        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_2.id,
        )
        corrupted_asset_2 = create_character_asset(
            character=self.character_1001, location=invalid_location_2
        )
        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_3.id,
        )
        corrupted_asset_3 = create_character_asset(
            character=self.character_1001, location=invalid_location_3
        )

        out = StringIO()

        # when
        with patch(
            MODULE_PATH + ".fix_invalid_locations",
            wraps=memberaudit_fix_locations.fix_invalid_locations,
        ) as spy_fix_corrupted_character_section:
            call_command(
                "memberaudit_fix_locations",
                "--noinput",
                "--batch-size-update",
                "1",
                stdout=out,
            )

            # then
            self.assertEqual(spy_fix_corrupted_character_section.call_count, 3)
            for corrupted_asset in [
                corrupted_asset_1,
                corrupted_asset_2,
                corrupted_asset_3,
            ]:
                corrupted_asset.refresh_from_db()
                self.assertEqual(
                    corrupted_asset.location.id, Location.LOCATION_UNKNOWN_ID
                )

    def test_should_fetch_locations_in_batches(self, *args, **kwargs):
        # given
        valid_location = create_location()
        invalid_location_1 = create_location()
        invalid_location_2 = create_location()
        invalid_location_3 = create_location()

        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_1.id,
        )
        create_character_asset(
            character=self.character_1001, location=invalid_location_1
        )
        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_2.id,
        )
        create_character_asset(
            character=self.character_1001, location=invalid_location_2
        )
        create_character_asset(
            character=self.character_1001,
            location=valid_location,
            item_id=invalid_location_3.id,
        )
        create_character_asset(
            character=self.character_1001, location=invalid_location_3
        )

        out = StringIO()

        # when
        with patch(
            MODULE_PATH + "._find_invalid_locations_chunk",
            wraps=memberaudit_fix_locations._find_invalid_locations_chunk,
        ) as spy_find_invalid_locations:
            call_command(
                "memberaudit_fix_locations",
                "--noinput",
                "--batch-size-fetch",
                "1",
                stdout=out,
            )

            # then
            self.assertEqual(spy_find_invalid_locations.call_count, 6 * 2)



from unittest.mock import patch

from django.contrib.admin.sites import AdminSite
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testdata_factories import UserFactory
from app_utils.testing import (
    create_authgroup,
    create_state,
    create_user_from_evecharacter,
)

from memberaudit.admin import (
    CharacterAdmin,
    ComplianceGroupDesignationAdmin,
    ComplianceGroupDesignationForm,
    SkillSetAdmin,
    SkillSetGroupAdmin,
)
from memberaudit.models import (
    Character,
    ComplianceGroupDesignation,
    SkillSet,
    SkillSetGroup,
)

from .testdata.factories import (
    create_character_update_status,
    create_compliance_group,
    create_skill_set,
)
from .testdata.load_entities import load_entities
from .testdata.load_eveuniverse import load_eveuniverse
from .utils import (
    add_memberaudit_character_to_user,
    create_memberaudit_character,
    create_user_from_evecharacter_with_access,
)

ADMIN_PATH = "memberaudit.admin"


class MockRequest(object):
    def __init__(self, user=None, post=None):
        self.user = user
        self.POST = post

    def get_full_path(self):
        return "/dummy-full-path"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestComplianceGroupDesignationAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = ComplianceGroupDesignationAdmin(
            model=ComplianceGroupDesignation, admin_site=AdminSite()
        )
        load_entities()
        cls.user = UserFactory(is_staff=True, is_superuser=True)

    def test_should_remove_deleted_compliance_group_from_users(self):
        # given
        compliance_group = create_compliance_group()
        obj = compliance_group.compliancegroupdesignation
        user_compliant, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user_compliant, 1001)
        user_compliant.groups.add(compliance_group)
        request = MockRequest(user=user_compliant)
        queryset = ComplianceGroupDesignation.objects.filter(pk=obj.pk)
        # when
        self.modeladmin.delete_queryset(request, queryset)
        # then
        self.assertFalse(ComplianceGroupDesignation.objects.filter(pk=obj.pk).exists())
        self.assertNotIn(compliance_group, user_compliant.groups.all())

    def test_should_add_group_to_compliant_users_when_created(self):
        # given
        compliance_group = create_authgroup(internal=True)
        user_compliant, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user_compliant, 1001)
        user_non_compliant, _ = create_user_from_evecharacter(
            1002, permissions=["memberaudit.basic_access"]
        )
        request = MockRequest(user=user_compliant)
        obj = ComplianceGroupDesignation(group=compliance_group)
        form = ComplianceGroupDesignationForm()
        # when
        self.modeladmin.save_model(request, obj, form, False)
        # then
        self.assertTrue(ComplianceGroupDesignation.objects.filter(pk=obj.pk).exists())
        self.assertIn(compliance_group, user_compliant.groups.all())
        self.assertNotIn(compliance_group, user_non_compliant.groups.all())

    def test_should_add_state_group_to_compliant_users_when_state_is_matching(self):
        # given
        member_corporation = EveCorporationInfo.objects.get(corporation_id=2001)
        my_state = create_state(member_corporations=[member_corporation], priority=200)
        compliance_group = create_authgroup(internal=True, states=[my_state])
        user_compliant, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user_compliant, 1001)
        request = MockRequest(user=user_compliant)
        obj = ComplianceGroupDesignation(group=compliance_group)
        form = ComplianceGroupDesignationForm()
        # when
        self.modeladmin.save_model(request, obj, form, False)
        # then
        self.assertTrue(ComplianceGroupDesignation.objects.filter(pk=obj.pk).exists())
        self.assertIn(compliance_group, user_compliant.groups.all())

    def test_should_not_add_new_state_group_to_compliant_user_when_state_not_matching(
        self,
    ):
        # given
        my_state = create_state(priority=200)
        compliance_group = create_authgroup(internal=True, states=[my_state])
        user_compliant, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user_compliant, 1001)
        request = MockRequest(user=user_compliant)
        obj = ComplianceGroupDesignation(group=compliance_group)
        form = ComplianceGroupDesignationForm()
        # when
        self.modeladmin.save_model(request, obj, form, False)
        # then
        self.assertNotIn(compliance_group, user_compliant.groups.all())

    def test_should_return_group_name(self):
        # given
        compliance_group = create_compliance_group(name="alpha")
        obj = compliance_group.compliancegroupdesignation
        # when
        result = self.modeladmin._group_name(obj)
        # then
        self.assertEqual(result, "alpha")

    def test_should_return_states_when_defined(self):
        # given
        my_state = create_state(priority=200, name="bravo")
        compliance_group = create_compliance_group(states=[my_state])
        obj = compliance_group.compliancegroupdesignation
        # when
        result = self.modeladmin._states(obj)
        # then
        self.assertIn("bravo", result)

    def test_should_return_empty_when_no_state_defined(self):
        # given
        compliance_group = create_compliance_group()
        obj = compliance_group.compliancegroupdesignation
        # when
        result = self.modeladmin._states(obj)
        # then
        self.assertEqual(result, "-")


class TestCharacterAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = CharacterAdmin(model=Character, admin_site=AdminSite())
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)
        cls.user = cls.character.eve_character.character_ownership.user

    def test_column_character(self):
        self.assertEqual(self.modeladmin._character(self.character), "Bruce Wayne")

    def test_column_main_normal(self):
        self.assertEqual(self.modeladmin._main(self.character), "Bruce Wayne")

    def test_column_main_no_main(self):
        # given
        character = create_memberaudit_character(1002)
        user = character.eve_character.character_ownership.user
        user.profile.main_character = None
        user.profile.save()
        # when
        self.assertIsNone(self.modeladmin._main(character))

    def test_column_state(self):
        self.assertEqual(self.modeladmin._state(self.character), "Guest")

    def test_column_organization_normal(self):
        self.assertEqual(
            self.modeladmin._organization(self.character), "Wayne Technologies [WYN]"
        )

    def test_column_organization_no_main(self):
        # given
        character = create_memberaudit_character(1002)
        user = character.eve_character.character_ownership.user
        user.profile.main_character = None
        user.profile.save()
        # when
        self.assertIsNone(self.modeladmin._organization(character))

    def test_column_missing_sections_none(self):
        # given
        for section in Character.UpdateSection:
            create_character_update_status(character=self.character, section=section)
        self.assertIsNone(self.modeladmin._missing_sections(self.character))

    def test_column_missing_sections_two_missing(self):
        # given
        sections = [
            obj
            for obj in Character.UpdateSection
            if obj is not Character.UpdateSection.ASSETS
            and obj is not Character.UpdateSection.CONTRACTS
        ]
        for section in sections:
            create_character_update_status(character=self.character, section=section)
        self.assertListEqual(
            self.modeladmin._missing_sections(self.character), ["assets", "contracts"]
        )

    @patch(ADMIN_PATH + ".CharacterAdmin.message_user")
    @patch(ADMIN_PATH + ".tasks.update_character")
    def test_should_update_characters(
        self, mock_task_update_character, mock_message_user
    ):
        # given
        request = MockRequest(user=self.user)
        queryset = Character.objects.all()
        # when
        self.modeladmin.update_characters(request, queryset)
        # then
        self.assertEqual(mock_task_update_character.apply_async.call_count, 1)
        self.assertTrue(mock_message_user.called)


@patch(ADMIN_PATH + ".CharacterAdmin.message_user")
@patch(ADMIN_PATH + ".tasks.delete_objects")
class TestCharacterDeleteCharactersAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = CharacterAdmin(model=Character, admin_site=AdminSite())
        load_eveuniverse()
        load_entities()
        cls.user = UserFactory(is_staff=True, is_superuser=True)

    def test_should_delete_characters_1(
        self, mock_task_delete_characters, mock_message_user
    ):
        # given
        factory = RequestFactory()
        request = factory.get(reverse("admin:memberaudit_character_changelist"))
        create_memberaudit_character(1001)
        queryset = Character.objects.all()
        # when
        response = self.modeladmin.delete_objects(request, queryset)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_delete_characters_2(
        self, mock_task_delete_characters, mock_message_user
    ):
        # given
        request = MockRequest(user=self.user, post="apply")
        create_memberaudit_character(1001)
        queryset = Character.objects.all()
        # when
        self.modeladmin.delete_objects(request, queryset)
        # then
        self.assertEqual(mock_task_delete_characters.apply_async.call_count, 1)
        self.assertTrue(mock_message_user.called)


@patch(ADMIN_PATH + ".tasks.update_characters_skill_checks")
class TestSkillSetAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.modeladmin = SkillSetAdmin(model=SkillSet, admin_site=AdminSite())
        load_eveuniverse()
        load_entities()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    def test_save_model(self, mock_update_characters_skill_checks):
        # given
        obj = SkillSet(name="Dummy")
        request = MockRequest(self.user)
        form = self.modeladmin.get_form(request)
        my_now = now()
        # when
        with patch(ADMIN_PATH + ".now", lambda: my_now):
            self.modeladmin.save_model(request, obj, form, True)
        # then
        obj_2: SkillSet = SkillSet.objects.get(name="Dummy")
        self.assertEqual(obj_2.last_modified_by, self.user)
        self.assertEqual(obj_2.last_modified_at, my_now)
        self.assertTrue(mock_update_characters_skill_checks.apply_async.called)

    def test_delete_model(self, mock_update_characters_skill_checks):
        # given
        obj = create_skill_set(name="Dummy")
        request = MockRequest(self.user)
        # when
        self.modeladmin.delete_model(request, obj)
        # then
        self.assertFalse(SkillSet.objects.filter(pk=obj.pk).exists())
        self.assertTrue(mock_update_characters_skill_checks.apply_async.called)

    # def test_ship_type_filter(self):
    #     class SkillSetAdminTest(SkillSetAdmin):
    #         list_filter = (SkillSetShipTypeFilter,)

    #     my_modeladmin = SkillSetAdminTest(SkillSet, AdminSite())

    #     ss_1 = SkillSet.objects.create(name="Set 1")
    #     ss_2 = SkillSet.objects.create(
    #         name="Set 2", ship_type=EveShipType.objects.get(id=603)
    #     )

    #     # Make sure the lookups are correct
    #     request = self.factory.get("/")
    #     request.user = self.user
    #     changelist = my_modeladmin.get_changelist_instance(request)
    #     filters = changelist.get_filters(request)
    #     filterspec = filters[0][0]
    #     expected = [("yes", "yes"), ("no", "no")]
    #     self.assertEqual(filterspec.lookup_choices, expected)

    #     # Make sure the correct queryset is returned
    #     request = self.factory.get("/", {"is_ship_type": "yes"})
    #     request.user = self.user
    #     changelist = my_modeladmin.get_changelist_instance(request)
    #     queryset = changelist.get_queryset(request)
    #     expected = {ss_2}
    #     self.assertSetEqual(set(queryset), expected)

    #     # Make sure the correct queryset is returned
    #     request = self.factory.get("/", {"is_ship_type": "no"})
    #     request.user = self.user
    #     changelist = my_modeladmin.get_changelist_instance(request)
    #     queryset = changelist.get_queryset(request)
    #     expected = {ss_1}
    #     self.assertSetEqual(set(queryset), expected)


class TestSkillSetSkillAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.user = UserFactory(is_staff=True, is_superuser=True)

    def test_should_create_new_skill_set_with_required_level_only(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            "/admin/memberaudit/skillset/add/",
            data={
                "name": "Bla Bla",
                "skills-TOTAL_FORMS": 1,
                "skills-INITIAL_FORMS": 0,
                "skills-MIN_NUM_FORMS": 1,
                "skills-MAX_NUM_FORMS": 1000,
                "skills-0-eve_type": 24311,
                "skills-0-required_level": 1,
            },
        )
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(SkillSet.objects.filter(name="Bla Bla").count(), 1)

    def test_should_create_new_skill_set_with_recommended_level_only(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            "/admin/memberaudit/skillset/add/",
            data={
                "name": "Bla Bla",
                "skills-TOTAL_FORMS": 1,
                "skills-INITIAL_FORMS": 0,
                "skills-MIN_NUM_FORMS": 1,
                "skills-MAX_NUM_FORMS": 1000,
                "skills-0-eve_type": 24311,
                "skills-0-recommended_level": 1,
            },
        )
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(SkillSet.objects.filter(name="Bla Bla").count(), 1)

    def test_should_raise_error_when_no_level_given(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            "/admin/memberaudit/skillset/add/",
            data={
                "name": "Bla Bla",
                "skills-TOTAL_FORMS": 1,
                "skills-INITIAL_FORMS": 0,
                "skills-MIN_NUM_FORMS": 1,
                "skills-MAX_NUM_FORMS": 1000,
                "skills-0-eve_type": 24311,
            },
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertIn("error", response.content.decode("utf-8"))


class TestSkillSetGroupAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.modeladmin = SkillSetGroupAdmin(model=SkillSetGroup, admin_site=AdminSite())
        load_eveuniverse()
        load_entities()
        cls.user = UserFactory(is_staff=True, is_superuser=True)

    def test_save_model(self):
        # given
        obj = SkillSetGroup(name="Dummy")
        request = MockRequest(self.user)
        form = self.modeladmin.get_form(request)
        my_now = now()
        # when
        with patch(ADMIN_PATH + ".now", lambda: my_now):
            self.modeladmin.save_model(request, obj, form, True)
        # then
        obj_2: SkillSetGroup = SkillSetGroup.objects.get(name="Dummy")
        self.assertEqual(obj_2.last_modified_by, self.user)
        self.assertEqual(obj_2.last_modified_at, my_now)


class TestCharacterAdminUi(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.user = UserFactory(is_staff=True, is_superuser=True)
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_render_list_view(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/admin/memberaudit/character/")
        # then
        self.assertEqual(response.status_code, 200)
        self.assertIn("Bruce Wayne", response.content.decode("utf-8"))

    def test_should_render_change_view(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get(
            f"/admin/memberaudit/character/{self.character.pk}/change/"
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertIn("Bruce Wayne", response.content.decode("utf-8"))



import csv
import datetime as dt
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch
from zipfile import ZipFile

from pytz import utc

from django.test import TestCase
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter

from memberaudit.core.data_exporters import (
    ContractExporter,
    ContractItemExporter,
    DataExporter,
    WalletJournalExporter,
    export_topic_to_archive,
    file_to_zip,
    topics_and_export_files,
)
from memberaudit.models import CharacterWalletJournalEntry
from memberaudit.tests.testdata.factories import (
    create_character,
    create_character_contract,
    create_character_contract_item,
    create_character_wallet_journal_entry,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.utils import create_memberaudit_character

MODULE_PATH = "memberaudit.core.data_exporters"


class TestExportTopicToArchive(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        cls.character = create_memberaudit_character(1001)

    def test_should_export_contract(self):
        with TemporaryDirectory() as tmpdirname:
            # given
            contract = create_character_contract(character=self.character)
            create_character_contract_item(contract=contract)
            # when
            result = export_topic_to_archive(
                topic="contract", destination_folder=tmpdirname
            )
            # then
            output_file = Path(result)
            self.assertTrue(output_file.exists())
            self.assertEqual("memberaudit_contract.zip", output_file.name)

    def test_should_export_contract_item(self):
        with TemporaryDirectory() as tmpdirname:
            # given
            contract = create_character_contract(character=self.character)
            create_character_contract_item(contract=contract)
            # when
            result = export_topic_to_archive(
                topic="contract-item", destination_folder=tmpdirname
            )
            # then
            output_file = Path(result)
            self.assertTrue(output_file.exists())
            self.assertEqual("memberaudit_contract-item.zip", output_file.name)

    def test_should_export_wallet_journal(self):
        with TemporaryDirectory() as tmpdirname:
            # given
            create_character_wallet_journal_entry(character=self.character)
            # when
            result = export_topic_to_archive(
                topic="wallet-journal", destination_folder=tmpdirname
            )
            # then
            output_file = Path(result)
            self.assertTrue(output_file.exists())
            self.assertEqual("memberaudit_wallet-journal.zip", output_file.name)

    def test_should_not_export_wallet_journal_when_no_data(self):
        with TemporaryDirectory() as tmpdirname:
            # when
            result = export_topic_to_archive(
                topic="wallet-journal", destination_folder=tmpdirname
            )
            # then
            self.assertEqual("", result)


class TestZipFile(TestCase):
    def test_should_zip_file_into_archive(self):
        with TemporaryDirectory() as tmpdirname_1, TemporaryDirectory() as tmpdirname_2:
            # given
            source_file = Path(tmpdirname_1) / "test.csv"
            with source_file.open("w") as fp:
                fp.write("test file")
            destination = Path(tmpdirname_2)
            # when
            zip_file = file_to_zip(source_file, destination)
            # then
            with ZipFile(zip_file, "r") as myzip:
                namelist = myzip.namelist()
            self.assertIn(source_file.name, namelist)


class NotTopicExporter(DataExporter):
    def format_obj(self, *args, **kwargs):
        return {}

    def get_queryset(self, *args, **kwargs):
        return None


class InvalidTopicExporter(NotTopicExporter):
    topic = "invalid_topic"


class TestDataExporter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.character_1121 = create_character(
            EveCharacter.objects.get(character_id=1121)
        )  # orphaned character

    def test_should_create_exporters(self):
        # given
        exporter_map = {
            "contract": ContractExporter,
            "contract-item": ContractItemExporter,
            "wallet-journal": WalletJournalExporter,
        }
        # when/then
        for topic, ExporterClass in exporter_map.items():
            obj = DataExporter.create_exporter(topic)
            # then
            self.assertIsInstance(obj, ExporterClass)

    def test_should_not_create_exporters(self):
        with self.assertRaises(ValueError):
            DataExporter.create_exporter("not-implemented")

    def test_should_return_topics(self):
        self.assertListEqual(
            DataExporter.topics(), ["contract", "contract-item", "wallet-journal"]
        )

    def test_can_not_init_exporter_without_topic(self):
        with self.assertRaises(ValueError):
            NotTopicExporter()

    def test_can_not_init_exporter_with_invalid_topic(self):
        with self.assertRaises(ValueError):
            InvalidTopicExporter()

    def test_should_return_title(self):
        # given
        exporter = DataExporter.create_exporter("wallet-journal")
        # when/then
        self.assertEqual(exporter.title, "Wallet Journal")

    def test_should_show_count(self):
        # given
        create_character_wallet_journal_entry(character=self.character_1001)
        exporter = DataExporter.create_exporter("wallet-journal")
        # when/then
        self.assertEqual(exporter.count(), 1)

    def test_should_have_data(self):
        # given
        create_character_wallet_journal_entry(character=self.character_1001)
        exporter = DataExporter.create_exporter("wallet-journal")
        # when/then
        self.assertTrue(exporter.has_data())

    def test_should_create_csv_file_for_contract(self):
        # given
        create_character_contract(character=self.character_1001, contract_id=42)
        create_character_contract(character=self.character_1121, contract_id=69)
        exporter = DataExporter.create_exporter("contract")
        # when
        data = self._write_to_file(exporter)
        # then
        self.assertEqual(len(data), 2)
        obj = data["Bruce Wayne"]
        self.assertEqual(obj["owner character"], "Bruce Wayne")
        self.assertEqual(obj["owner corporation"], "Wayne Technologies")
        self.assertEqual(obj["contract id"], "42")
        # TODO: test all properties and all contract types

    def test_should_create_csv_file_for_contract_item(self):
        # given
        contract = create_character_contract(character=self.character_1001)
        create_character_contract_item(contract=contract, record_id=12)
        exporter = DataExporter.create_exporter("contract-item")
        # when
        data = self._write_to_file(exporter, "record id")
        # then
        self.assertEqual(len(data), 1)
        obj = data["12"]
        self.assertEqual(obj["contract pk"], str(contract.pk))
        self.assertEqual(obj["record id"], "12")
        self.assertEqual(obj["type"], "Merlin")
        self.assertEqual(obj["quantity"], "1")
        # TODO: test all properties and all contract types

    def test_should_create_csv_file_for_wallet_journal(self):
        # given
        create_character_wallet_journal_entry(
            amount=1000000.0,
            balance=20000000.0,
            character=self.character_1001,
            context_id=1002,
            context_id_type=CharacterWalletJournalEntry.CONTEXT_ID_TYPE_CHARACTER_ID,
            date=dt.datetime(2021, 12, 1, 12, 30, tzinfo=utc),
            description="test description",
            entry_id=42,
            first_party_id=1001,
            ref_type="player_donation",
            reason="test reason",
            second_party_id=1002,
            tax=0.05,
            tax_receiver_id=1101,
        )
        create_character_wallet_journal_entry(
            amount=1000000.0,
            balance=20000000.0,
            character=self.character_1121,
            context_id=1002,
            context_id_type=CharacterWalletJournalEntry.CONTEXT_ID_TYPE_CHARACTER_ID,
            date=dt.datetime(2021, 12, 1, 12, 30, tzinfo=utc),
            description="test description",
            entry_id=69,
            first_party_id=1001,
            ref_type="player_donation",
            reason="test reason",
            second_party_id=1002,
            tax=0.05,
            tax_receiver_id=1101,
        )
        exporter = DataExporter.create_exporter("wallet-journal")
        # when
        data = self._write_to_file(exporter)
        # then
        self.assertEqual(len(data), 2)
        obj = data["Bruce Wayne"]
        self.assertEqual(obj["owner character"], "Bruce Wayne")
        self.assertEqual(obj["owner corporation"], "Wayne Technologies")
        self.assertEqual(obj["amount"], "1000000.0")
        self.assertEqual(obj["balance"], "20000000.0")
        self.assertEqual(obj["date"], "2021-12-01 12:30:00")
        self.assertEqual(obj["description"], "test description")
        self.assertEqual(obj["entry id"], "42")
        self.assertEqual(obj["first party"], "Bruce Wayne")
        self.assertEqual(obj["ref type"], "Player Donation")
        self.assertEqual(obj["second party"], "Clark Kent")
        self.assertEqual(obj["reason"], "test reason")
        self.assertEqual(obj["context_id"], "1002")
        self.assertEqual(obj["context_id_type"], "character ID")
        self.assertEqual(obj["tax"], "0.05")
        self.assertEqual(obj["tax_receiver"], "Lex Luther")

    def _write_to_file(self, exporter, key="owner character") -> dict:
        with TemporaryDirectory() as tmpdirname:
            output_file = exporter.write_to_file(tmpdirname)
            self.assertTrue(output_file.exists())
            with output_file.open("r") as csv_file:
                reader = csv.DictReader(csv_file)
                return {row[key]: row for row in reader}


class TestTopicsAndExportFiles(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        cls.character = create_memberaudit_character(1001)

    @patch(MODULE_PATH + ".MEMBERAUDIT_DATA_EXPORT_MIN_UPDATE_AGE", 60)
    def test_should_return_correct_list(self):
        with TemporaryDirectory() as tmpdirname:
            # given
            contract = create_character_contract(character=self.character)
            create_character_contract_item(contract=contract)
            contract_item_file = Path(tmpdirname) / "memberaudit_contract-item.zip"
            contract_item_file.touch()
            new_ts = (now() - dt.timedelta(minutes=10)).timestamp()
            os.utime(contract_item_file, (new_ts, new_ts))
            contract_file = Path(tmpdirname) / "memberaudit_contract.zip"
            contract_file.touch()
            contract_file_dt = now() - dt.timedelta(minutes=61)
            new_ts = contract_file_dt.timestamp()
            os.utime(contract_file, (new_ts, new_ts))
            wrong_file = Path(tmpdirname) / "memberaudit.zip"
            wrong_file.touch()
            # when
            result = topics_and_export_files(tmpdirname)
            result_2 = {obj["value"]: obj for obj in result}
            # then
            self.assertListEqual(
                list(result_2.keys()), ["contract", "contract-item", "wallet-journal"]
            )
            contract = result_2["contract"]
            self.assertEqual(contract["value"], "contract")
            self.assertEqual(contract["title"], "Contract")
            self.assertEqual(contract["rows"], 1)
            self.assertEqual(contract["last_updated_at"], contract_file_dt)
            self.assertTrue(contract["has_file"])
            self.assertTrue(contract["update_allowed"])
            contract_item = result_2["contract-item"]
            self.assertTrue(contract_item["has_file"])
            self.assertAlmostEqual(
                contract_item["last_updated_at"],
                now(),
                delta=dt.timedelta(minutes=10, seconds=10),
            )
            self.assertFalse(contract_item["update_allowed"])
            wallet_journal = result_2["wallet-journal"]
            self.assertFalse(wallet_journal["has_file"])
            self.assertEqual(wallet_journal["rows"], 0)



import datetime as dt
from unittest.mock import patch

from django.test import override_settings
from django.utils.dateparse import parse_datetime
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveSolarSystem, EveType

from app_utils.esi_testing import (
    BravadoOperationStub,
    EsiClientStub,
    EsiEndpoint,
    build_http_error,
)
from app_utils.testing import NoSocketsTestCase

from memberaudit.models import (
    Character,
    CharacterOnlineStatus,
    CharacterPlanet,
    CharacterRole,
    CharacterShip,
    CharacterSkill,
    CharacterSkillqueueEntry,
    CharacterSkillSetCheck,
    CharacterStanding,
    CharacterWalletBalance,
    CharacterWalletJournalEntry,
    CharacterWalletTransaction,
    Location,
)
from memberaudit.tests.constants import EvePlanetIds, EveSolarSystemIds
from memberaudit.tests.testdata.constants import EveTypeId
from memberaudit.tests.testdata.esi_client_stub import esi_client_stub
from memberaudit.tests.testdata.factories import (
    create_character_mining_ledger_entry,
    create_character_planet,
    create_character_role,
    create_character_ship,
    create_character_skill,
    create_character_skill_set_check,
    create_character_standing,
    create_character_title,
    create_character_wallet_journal_entry,
    create_skill_set,
    create_skill_set_group,
    create_skill_set_skill,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import create_memberaudit_character

MODELS_PATH = "memberaudit.models"
MANAGERS_PATH = "memberaudit.managers.character_sections_3"


@patch(MANAGERS_PATH + ".esi")
class TestCharacterMiningLedgerManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.endpoints = [
            EsiEndpoint(
                "Industry",
                "get_characters_character_id_mining",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {
                            "date": "2017-09-19",
                            "quantity": 7004,
                            "solar_system_id": 30002537,
                            "type_id": 17471,
                        },
                        {
                            "date": "2017-09-18",
                            "quantity": 5199,
                            "solar_system_id": 30002537,
                            "type_id": 17471,
                        },
                    ]
                },
            ),
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(cls.endpoints)

    def test_should_add_new_entry_from_scratch(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        # when
        self.character_1001.update_mining_ledger()
        # then
        self.assertEqual(self.character_1001.mining_ledger.count(), 2)
        obj = self.character_1001.mining_ledger.first()
        self.assertEqual(obj.date, dt.date(2017, 9, 19))
        self.assertEqual(obj.eve_type, EveType.objects.get(name="Dense Veldspar"))
        self.assertEqual(
            obj.eve_solar_system, EveSolarSystem.objects.get(name="Amamake")
        )
        self.assertEqual(obj.quantity, 7004)

    def test_should_update_existing_entries(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        create_character_mining_ledger_entry(
            character=self.character_1001,
            date=dt.date(2017, 9, 19),
            eve_solar_system=EveSolarSystem.objects.get(name="Amamake"),
            eve_type=EveType.objects.get(name="Dense Veldspar"),
            quantity=5,
        )
        # when
        self.character_1001.update_mining_ledger()
        # then
        self.assertEqual(self.character_1001.mining_ledger.count(), 2)
        obj = self.character_1001.mining_ledger.get(
            date=dt.date(2017, 9, 19),
            eve_solar_system=EveSolarSystem.objects.get(name="Amamake"),
            eve_type=EveType.objects.get(name="Dense Veldspar"),
        )
        self.assertEqual(obj.quantity, 7004)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterOnlineStatusManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_update_online_status(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterOnlineStatus.objects.update_or_create_esi(self.character)
        # then
        self.assertEqual(
            self.character.online_status.last_login,
            parse_datetime("2017-01-02T03:04:05Z"),
        )
        self.assertEqual(
            self.character.online_status.last_logout,
            parse_datetime("2017-01-02T04:05:06Z"),
        )
        self.assertEqual(self.character.online_status.logins, 9001)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterPlanetManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.endpoints = [
            EsiEndpoint(
                "Planetary_Interaction",
                "get_characters_character_id_planets",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {
                            "last_update": "2016-11-28T16:42:51Z",
                            "num_pins": 1,
                            "owner_id": 1001,
                            "planet_id": EvePlanetIds.AMAMAKE_I.value,
                            "planet_type": "barren",
                            "solar_system_id": EveSolarSystemIds.AMAMAKE.value,
                            "upgrade_level": 0,
                        },
                        {
                            "last_update": "2016-10-29T15:41:21Z",
                            "num_pins": 3,
                            "owner_id": 1001,
                            "planet_id": EvePlanetIds.AMAMAKE_II.value,
                            "planet_type": "barren",
                            "solar_system_id": EveSolarSystemIds.AMAMAKE.value,
                            "upgrade_level": 2,
                        },
                    ]
                },
            ),
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(cls.endpoints)

    def test_should_create_new_planets_from_scratch(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub

        # when
        self.character_1001.update_planets()

        # then
        self.assertEqual(self.character_1001.planets.count(), 2)

        obj: CharacterPlanet = self.character_1001.planets.get(
            eve_planet_id=EvePlanetIds.AMAMAKE_I
        )
        self.assertIsInstance(obj.last_update_at, dt.datetime)
        self.assertEqual(obj.num_pins, 1)
        self.assertEqual(obj.upgrade_level, 0)

        obj: CharacterPlanet = self.character_1001.planets.get(
            eve_planet_id=EvePlanetIds.AMAMAKE_II
        )
        self.assertIsInstance(obj.last_update_at, dt.datetime)
        self.assertEqual(obj.num_pins, 3)
        self.assertEqual(obj.upgrade_level, 2)

    def test_should_update_existing_planets(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        obj_1 = create_character_planet(
            character=self.character_1001,
            eve_planet_id=EvePlanetIds.AMAMAKE_I,
            num_pins=2,
            upgrade_level=3,
        )
        create_character_planet(
            character=self.character_1001,
            eve_planet_id=EvePlanetIds.AMAMAKE_II,
            num_pins=3,
            upgrade_level=2,
            last_update_at=parse_datetime("2016-10-29T15:41:21Z"),
        )

        with patch(
            MANAGERS_PATH + ".CharacterPlanetManager.bulk_update",
            wraps=CharacterPlanet.objects.filter(
                character=self.character_1001
            ).bulk_update,
        ) as mock_bulk_update:
            # when
            self.character_1001.update_planets()

            # then
            self.assertEqual(self.character_1001.planets.count(), 2)

            obj_1.refresh_from_db()
            self.assertIsInstance(obj_1.last_update_at, dt.datetime)
            self.assertEqual(obj_1.num_pins, 1)
            self.assertEqual(obj_1.upgrade_level, 0)

            # then only the modified planet was updated
            updated_obj_ids = {o.id for o in mock_bulk_update.call_args.kwargs["objs"]}
            self.assertSetEqual(updated_obj_ids, {obj_1.id})

    def test_should_remove_obsolete_planets(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        create_character_planet(
            character=self.character_1001, eve_planet_id=EvePlanetIds.AMAMAKE_IV
        )

        # when
        self.character_1001.update_planets()

        # then
        current_planet_ids = set(
            self.character_1001.planets.values_list("eve_planet_id", flat=True)
        )
        self.assertSetEqual(
            current_planet_ids, {EvePlanetIds.AMAMAKE_I, EvePlanetIds.AMAMAKE_II}
        )

    def test_should_remove_all_when_no_esi_data(self, mock_esi):
        # given
        create_character_planet(
            character=self.character_1001, eve_planet_id=EvePlanetIds.AMAMAKE_IV
        )
        endpoints = [
            EsiEndpoint(
                "Planetary_Interaction",
                "get_characters_character_id_planets",
                "character_id",
                needs_token=True,
                data={"1001": []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)

        # when
        self.character_1001.update_planets()

        # then
        self.assertEqual(self.character_1001.planets.count(), 0)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterRolesManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        with patch(MODELS_PATH + ".characters.MEMBERAUDIT_FEATURE_ROLES_ENABLED", True):
            cls.character_1001 = create_memberaudit_character(1001)

    def test_should_add_new_role(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_roles",
                "character_id",
                needs_token=True,
                data={
                    "1001": {
                        "roles": ["Station_Manager"],
                    }
                },
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        self.character_1001.update_roles()
        # then
        self.assertEqual(self.character_1001.roles.count(), 1)
        obj = self.character_1001.roles.first()
        self.assertEqual(obj.role, CharacterRole.Role.STATION_MANAGER)
        self.assertEqual(obj.location, CharacterRole.Location.UNIVERSAL)

    def test_should_update_existing_entries(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_roles",
                "character_id",
                needs_token=True,
                data={
                    "1001": {
                        "roles": ["Station_Manager"],
                    }
                },
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        create_character_role(
            character=self.character_1001, role=CharacterRole.Role.SECURITY_OFFICER
        )
        # when
        self.character_1001.update_roles()
        # then
        obj = self.character_1001.roles.first()
        self.assertEqual(obj.role, CharacterRole.Role.STATION_MANAGER)
        self.assertEqual(obj.location, CharacterRole.Location.UNIVERSAL)

    def test_should_keep_existing_entries(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_roles",
                "character_id",
                needs_token=True,
                data={
                    "1001": {
                        "roles": ["Station_Manager"],
                    }
                },
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        create_character_role(
            character=self.character_1001, role=CharacterRole.Role.STATION_MANAGER
        )
        # when
        self.character_1001.update_roles()
        # then
        obj = self.character_1001.roles.first()
        self.assertEqual(obj.role, CharacterRole.Role.STATION_MANAGER)
        self.assertEqual(obj.location, CharacterRole.Location.UNIVERSAL)

    # this catches #159
    def test_should_handle_unknown_roles_gracefully(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_roles",
                "character_id",
                needs_token=True,
                data={
                    "1001": {
                        "roles": ["Station_Manager", "My_Unknown_Role"],
                    }
                },
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        self.character_1001.update_roles()
        # then
        self.assertEqual(self.character_1001.roles.count(), 1)
        obj = self.character_1001.roles.first()
        self.assertEqual(obj.role, CharacterRole.Role.STATION_MANAGER)
        self.assertEqual(obj.location, CharacterRole.Location.UNIVERSAL)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterShipManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    def test_should_update_all_fields(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterShip.objects.update_or_create_esi(self.character_1001)
        # then
        self.assertEqual(self.character_1001.ship.item_id, 1000000016991)
        self.assertEqual(self.character_1001.ship.eve_type_id, 603)
        self.assertEqual(self.character_1001.ship.name, "Shooter Boy")

    def test_should_ignore_error_500(self, mock_esi):
        # given
        error_500 = build_http_error(
            500, '{"error":"Undefined 404 response. Original message: Ship not found"}'
        )
        mock_esi.client.Location.get_characters_character_id_ship.side_effect = (
            error_500
        )
        create_character_ship(
            character=self.character_1001,
            eve_type_id=603,
            item_id=1000000016991,
            name="Shooter Boy",
        )
        # when
        CharacterShip.objects.update_or_create_esi(self.character_1001)
        # then
        self.character_1001.refresh_from_db()
        self.assertEqual(self.character_1001.ship.eve_type_id, 603)
        self.assertEqual(self.character_1001.ship.name, "Shooter Boy")
        self.assertEqual(self.character_1001.ship.item_id, 1000000016991)

    def test_should_remove_ship_when_esi_returns_empty_response(self, mock_esi):
        # given
        mock_esi.client.Location.get_characters_character_id_ship.return_value = (
            BravadoOperationStub(data={})
        )
        create_character_ship(character=self.character_1001)
        # when
        CharacterShip.objects.update_or_create_esi(self.character_1001)
        # then
        self.assertFalse(
            CharacterShip.objects.filter(character=self.character_1001).exists()
        )


@patch(MANAGERS_PATH + ".esi")
class TestCharacterSkillManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    def test_can_create_new_skills(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterSkill.objects.update_or_create_esi(character=self.character_1001)
        # then
        self.assertEqual(self.character_1001.skillpoints.total, 30_000)
        self.assertEqual(self.character_1001.skillpoints.unallocated, 1_000)
        self.assertSetEqual(
            set(self.character_1001.skills.values_list("eve_type_id", flat=True)),
            {EveTypeId.AMARR_CARRIER, EveTypeId.CALDARI_CARRIER},
        )
        skill = self.character_1001.skills.get(eve_type_id=EveTypeId.AMARR_CARRIER)
        self.assertEqual(skill.active_skill_level, 3)
        self.assertEqual(skill.skillpoints_in_skill, 20_000)
        self.assertEqual(skill.trained_skill_level, 4)

        skill = self.character_1001.skills.get(eve_type_id=EveTypeId.CALDARI_CARRIER)
        self.assertEqual(skill.active_skill_level, 1)
        self.assertEqual(skill.skillpoints_in_skill, 10_000)
        self.assertEqual(skill.trained_skill_level, 1)

    def test_caen_update_existing_skills(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_skill(
            character=self.character_1001,
            eve_type=EveType.objects.get(id=EveTypeId.AMARR_CARRIER),
            active_skill_level=1,
            skillpoints_in_skill=1,
            trained_skill_level=1,
        )
        # when
        CharacterSkill.objects.update_or_create_esi(character=self.character_1001)
        # then
        self.assertEqual(self.character_1001.skills.count(), 2)
        skill = self.character_1001.skills.get(eve_type_id=EveTypeId.AMARR_CARRIER)
        self.assertEqual(skill.active_skill_level, 3)
        self.assertEqual(skill.skillpoints_in_skill, 20_000)
        self.assertEqual(skill.trained_skill_level, 4)

    def test_can_delete_obsolete_skills(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_skill(
            character=self.character_1001,
            eve_type=EveType.objects.get(id=EveTypeId.MERLIN),
            active_skill_level=1,
            skillpoints_in_skill=1,
            trained_skill_level=1,
        )
        # when
        CharacterSkill.objects.update_or_create_esi(character=self.character_1001)
        # then
        self.assertSetEqual(
            set(self.character_1001.skills.values_list("eve_type_id", flat=True)),
            {EveTypeId.AMARR_CARRIER, EveTypeId.CALDARI_CARRIER},
        )

    def test_update_skills_4(self, mock_esi):
        """when ESI info has not changed, then do not update local data"""
        # given
        mock_esi.client = esi_client_stub
        self.character_1001.reset_update_section(Character.UpdateSection.SKILLS)
        self.character_1001.update_skills()
        skill = self.character_1001.skills.get(eve_type_id=EveTypeId.AMARR_CARRIER)
        skill.active_skill_level = 4
        skill.save()
        # when
        CharacterSkill.objects.update_or_create_esi(character=self.character_1001)
        # then
        skill.refresh_from_db()
        self.assertEqual(skill.active_skill_level, 4)

    def test_update_skills_5(self, mock_esi):
        """when ESI info has not changed and update forced, then update local data"""
        # given
        mock_esi.client = esi_client_stub
        self.character_1001.reset_update_section(Character.UpdateSection.SKILLS)
        # when
        CharacterSkill.objects.update_or_create_esi(character=self.character_1001)
        # then
        skill = self.character_1001.skills.get(eve_type_id=EveTypeId.AMARR_CARRIER)
        skill.active_skill_level = 4
        skill.save()
        self.character_1001.update_skills(force_update=True)
        skill = self.character_1001.skills.get(eve_type_id=EveTypeId.AMARR_CARRIER)
        self.assertEqual(skill.active_skill_level, 3)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterSkillqueueEntryManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    def test_can_create_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterSkillqueueEntry.objects.update_or_create_esi(self.character_1001)
        # then
        self.assertEqual(self.character_1001.skillqueue.count(), 3)

        entry = self.character_1001.skillqueue.get(queue_position=0)
        self.assertEqual(
            entry.eve_type, EveType.objects.get(id=EveTypeId.AMARR_CARRIER)
        )
        self.assertEqual(entry.finish_date, parse_datetime("2016-06-29T10:47:00Z"))
        self.assertEqual(entry.finished_level, 3)
        self.assertEqual(entry.start_date, parse_datetime("2016-06-29T10:46:00Z"))

        entry = self.character_1001.skillqueue.get(queue_position=1)
        self.assertEqual(
            entry.eve_type, EveType.objects.get(id=EveTypeId.CALDARI_CARRIER)
        )
        self.assertEqual(entry.finish_date, parse_datetime("2016-07-15T10:47:00Z"))
        self.assertEqual(entry.finished_level, 4)
        self.assertEqual(entry.level_end_sp, 1000)
        self.assertEqual(entry.level_start_sp, 100)
        self.assertEqual(entry.start_date, parse_datetime("2016-06-29T10:47:00Z"))
        self.assertEqual(entry.training_start_sp, 50)

        entry = self.character_1001.skillqueue.get(queue_position=2)
        self.assertEqual(
            entry.eve_type, EveType.objects.get(id=EveTypeId.CALDARI_CARRIER)
        )
        self.assertEqual(entry.finished_level, 5)

    def test_can_update_existing_queue(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        self.character_1001.skillqueue.create(
            queue_position=0,
            eve_type=EveType.objects.get(id=EveTypeId.AMARR_CARRIER),
            finish_date=now() + dt.timedelta(days=1),
            finished_level=4,
            start_date=now() - dt.timedelta(days=1),
        )
        # when
        CharacterSkillqueueEntry.objects.update_or_create_esi(self.character_1001)
        # then
        self.assertEqual(self.character_1001.skillqueue.count(), 3)

        entry = self.character_1001.skillqueue.get(queue_position=0)
        self.assertEqual(
            entry.eve_type, EveType.objects.get(id=EveTypeId.AMARR_CARRIER)
        )
        self.assertEqual(entry.finish_date, parse_datetime("2016-06-29T10:47:00Z"))
        self.assertEqual(entry.finished_level, 3)
        self.assertEqual(entry.start_date, parse_datetime("2016-06-29T10:46:00Z"))

    def test_should_skip_update_when_no_change(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        CharacterSkillqueueEntry.objects.update_or_create_esi(self.character_1001)
        entry = self.character_1001.skillqueue.get(queue_position=0)
        entry.finished_level = 4
        entry.save()
        # when
        CharacterSkillqueueEntry.objects.update_or_create_esi(self.character_1001)
        # then
        entry = self.character_1001.skillqueue.get(queue_position=0)
        self.assertEqual(entry.finished_level, 4)

    def test_should_always_update_when_forced(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        CharacterSkillqueueEntry.objects.update_or_create_esi(self.character_1001)
        entry = self.character_1001.skillqueue.get(queue_position=0)
        entry.finished_level = 4
        entry.save()
        # when
        CharacterSkillqueueEntry.objects.update_or_create_esi(
            self.character_1001, force_update=True
        )
        # then
        entry = self.character_1001.skillqueue.get(queue_position=0)
        self.assertEqual(entry.finished_level, 3)

    def test_should_remove_existing_entries(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Skills",
                "get_characters_character_id_skillqueue",
                "character_id",
                needs_token=True,
                data={"1001": []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        self.character_1001.skillqueue.create(
            queue_position=0,
            eve_type=EveType.objects.get(id=EveTypeId.AMARR_CARRIER),
            finish_date=now() + dt.timedelta(days=1),
            finished_level=4,
            start_date=now() - dt.timedelta(days=1),
        )
        # when
        self.character_1001.update_skill_queue()
        # then
        self.assertEqual(self.character_1001.skillqueue.count(), 0)


class TestCharacterSkillSetCheckManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)
        cls.user = cls.character.eve_character.character_ownership.user
        # amarr carrier skill set
        cls.amarr_carrier_skill_type = EveType.objects.get(name="Amarr Carrier")
        cls.amarr_carrier_skill_set = create_skill_set()
        cls.amarr_carrier_skill_set_skill = create_skill_set_skill(
            skill_set=cls.amarr_carrier_skill_set,
            eve_type=cls.amarr_carrier_skill_type,
            required_level=3,
            recommended_level=5,
        )
        # caldari carrier skill set
        cls.caldari_carrier_skill_type = EveType.objects.get(name="Caldari Carrier")
        cls.caldari_carrier_skill_set = create_skill_set()
        cls.caldari_carrier_skill_set_skill = create_skill_set_skill(
            skill_set=cls.caldari_carrier_skill_set,
            eve_type=cls.caldari_carrier_skill_type,
            required_level=3,
            recommended_level=5,
        )

    def test_should_record_character_has_all_required_but_missing_recommended_skills(
        self,
    ):
        # given
        create_character_skill(self.character, eve_type=self.amarr_carrier_skill_type)
        # when
        self.character.update_skill_sets()
        # then
        obj: CharacterSkillSetCheck = self.character.skill_set_checks.filter(
            skill_set=self.amarr_carrier_skill_set
        ).first()
        self.assertTrue(obj.can_fly)
        self.assertEqual(obj.failed_required_skills.count(), 0)
        self.assertIn(
            self.amarr_carrier_skill_set_skill, obj.failed_recommended_skills.all()
        )
        obj: CharacterSkillSetCheck = self.character.skill_set_checks.filter(
            skill_set=self.caldari_carrier_skill_set
        ).first()
        self.assertFalse(obj.can_fly)

    def test_should_record_character_is_missing_all_skills(self):
        # given
        create_character_skill(
            self.character, eve_type=self.amarr_carrier_skill_type, active_skill_level=1
        )
        # when
        self.character.update_skill_sets()
        # then
        obj: CharacterSkillSetCheck = self.character.skill_set_checks.filter(
            skill_set=self.amarr_carrier_skill_set
        ).first()
        self.assertFalse(obj.can_fly)
        self.assertIn(
            self.amarr_carrier_skill_set_skill, obj.failed_required_skills.all()
        )
        self.assertIn(
            self.amarr_carrier_skill_set_skill, obj.failed_recommended_skills.all()
        )
        obj: CharacterSkillSetCheck = self.character.skill_set_checks.filter(
            skill_set=self.caldari_carrier_skill_set
        ).first()
        self.assertFalse(obj.can_fly)

    def test_should_update_existing_skill_set_check(self):
        # given
        create_character_skill(
            character=self.character,
            eve_type=self.amarr_carrier_skill_type,
            active_skill_level=5,
        )
        skill_set_check = create_character_skill_set_check(
            character=self.character, skill_set=self.amarr_carrier_skill_set
        )
        skill_set_check.failed_required_skills.add(self.amarr_carrier_skill_set_skill)
        skill_set_check.failed_recommended_skills.add(
            self.amarr_carrier_skill_set_skill
        )
        # when
        self.character.update_skill_sets()
        # then
        obj: CharacterSkillSetCheck = self.character.skill_set_checks.filter(
            skill_set=self.amarr_carrier_skill_set
        ).first()
        self.assertTrue(obj.can_fly)
        self.assertEqual(obj.failed_required_skills.count(), 0)
        self.assertEqual(obj.failed_recommended_skills.count(), 0)
        obj: CharacterSkillSetCheck = self.character.skill_set_checks.filter(
            skill_set=self.caldari_carrier_skill_set
        ).first()
        self.assertFalse(obj.can_fly)


class TestCharacterUpdateSkillSets(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)
        cls.amarr_carrier_skill_type = EveType.objects.get(id=24311)
        cls.caldari_carrier_skill_type = EveType.objects.get(id=24312)

    def test_has_all_skills(self):
        # given
        create_character_skill(
            character=self.character,
            eve_type=self.amarr_carrier_skill_type,
            active_skill_level=5,
            skillpoints_in_skill=10,
            trained_skill_level=5,
        )
        create_character_skill(
            character=self.character,
            eve_type=self.caldari_carrier_skill_type,
            active_skill_level=5,
            skillpoints_in_skill=10,
            trained_skill_level=5,
        )
        skill_set = create_skill_set()
        create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.amarr_carrier_skill_type,
            required_level=5,
        )
        create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.caldari_carrier_skill_type,
            required_level=3,
        )
        skill_set_group = create_skill_set_group()
        skill_set_group.skill_sets.add(skill_set)

        # when
        result = self.character.update_skill_sets()

        # then
        self.assertTrue(result.is_updated)
        self.assertEqual(self.character.skill_set_checks.count(), 1)
        first = self.character.skill_set_checks.first()
        self.assertEqual(first.skill_set.pk, skill_set.pk)
        self.assertEqual(first.failed_required_skills.count(), 0)

    def test_one_skill_below(self):
        # given
        create_character_skill(
            character=self.character,
            eve_type=self.amarr_carrier_skill_type,
            active_skill_level=5,
            skillpoints_in_skill=10,
            trained_skill_level=5,
        )
        create_character_skill(
            character=self.character,
            eve_type=self.caldari_carrier_skill_type,
            active_skill_level=2,
            skillpoints_in_skill=10,
            trained_skill_level=5,
        )
        skill_set = create_skill_set()
        create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.amarr_carrier_skill_type,
            required_level=5,
        )
        skill_2 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.caldari_carrier_skill_type,
            required_level=3,
        )
        skill_set_group = create_skill_set_group()
        skill_set_group.skill_sets.add(skill_set)

        # when
        result = self.character.update_skill_sets()

        # then
        self.assertTrue(result.is_updated)
        self.assertEqual(self.character.skill_set_checks.count(), 1)
        first = self.character.skill_set_checks.first()
        self.assertEqual(first.skill_set.pk, skill_set.pk)
        required_skill_pks = {obj.pk for obj in first.failed_required_skills.all()}
        self.assertEqual(required_skill_pks, {skill_2.pk})

    def test_misses_one_skill(self):
        # given
        create_character_skill(
            character=self.character,
            eve_type=self.amarr_carrier_skill_type,
            active_skill_level=5,
            skillpoints_in_skill=10,
            trained_skill_level=5,
        )
        skill_set = create_skill_set()
        create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.amarr_carrier_skill_type,
            required_level=5,
        )
        skill_2 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.caldari_carrier_skill_type,
            required_level=3,
        )
        skill_set_group = create_skill_set_group()
        skill_set_group.skill_sets.add(skill_set)

        # when
        result = self.character.update_skill_sets()

        # then
        self.assertTrue(result.is_updated)

        self.assertEqual(self.character.skill_set_checks.count(), 1)
        first = self.character.skill_set_checks.first()
        self.assertEqual(first.skill_set.pk, skill_set.pk)
        required_skill_pks = {obj.pk for obj in first.failed_required_skills.all()}
        self.assertSetEqual(required_skill_pks, {skill_2.pk})

    def test_passed_required_and_misses_recommended_skill(self):
        # given
        create_character_skill(
            character=self.character,
            eve_type=self.amarr_carrier_skill_type,
            active_skill_level=4,
            skillpoints_in_skill=10,
            trained_skill_level=4,
        )
        skill_set = create_skill_set()
        skill_1 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.amarr_carrier_skill_type,
            required_level=3,
            recommended_level=5,
        )

        # when
        result = self.character.update_skill_sets()

        # then
        self.assertTrue(result.is_updated)

        self.assertEqual(self.character.skill_set_checks.count(), 1)
        first = self.character.skill_set_checks.first()
        self.assertEqual(first.skill_set.pk, skill_set.pk)
        required_skill_pks = {obj.pk for obj in first.failed_required_skills.all()}
        self.assertSetEqual(required_skill_pks, set())
        recommended_skill_pks = {
            obj.pk for obj in first.failed_recommended_skills.all()
        }
        self.assertSetEqual(recommended_skill_pks, {skill_1.pk})

    def test_misses_recommended_skill_only(self):
        # given
        create_character_skill(
            character=self.character,
            eve_type=self.amarr_carrier_skill_type,
            active_skill_level=4,
            skillpoints_in_skill=10,
            trained_skill_level=4,
        )
        skill_set = create_skill_set()
        skill_1 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.amarr_carrier_skill_type,
            recommended_level=5,
        )

        # when
        result = self.character.update_skill_sets()

        # then
        self.assertTrue(result.is_updated)

        self.assertEqual(self.character.skill_set_checks.count(), 1)
        first = self.character.skill_set_checks.first()
        self.assertEqual(first.skill_set.pk, skill_set.pk)
        required_skill_pks = {obj.pk for obj in first.failed_required_skills.all()}
        self.assertSetEqual(required_skill_pks, set())
        recommended_skill_pks = {
            obj.pk for obj in first.failed_recommended_skills.all()
        }
        self.assertSetEqual(recommended_skill_pks, {skill_1.pk})

    def test_misses_all_skills(self):
        # given
        skill_set = create_skill_set()
        skill_1 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.amarr_carrier_skill_type,
            required_level=5,
        )
        skill_2 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.caldari_carrier_skill_type,
            required_level=3,
        )
        skill_set_group = create_skill_set_group()
        skill_set_group.skill_sets.add(skill_set)

        # when
        result = self.character.update_skill_sets()

        # then
        self.assertTrue(result.is_updated)

        self.assertEqual(self.character.skill_set_checks.count(), 1)
        first = self.character.skill_set_checks.first()
        self.assertEqual(first.skill_set.pk, skill_set.pk)
        skill_pks = {obj.pk for obj in first.failed_required_skills.all()}
        self.assertSetEqual(skill_pks, {skill_1.pk, skill_2.pk})

    def test_does_not_require_doctrine_definition(self):
        # given
        skill_set = create_skill_set()
        skill_1 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.amarr_carrier_skill_type,
            required_level=5,
        )
        skill_2 = create_skill_set_skill(
            skill_set=skill_set,
            eve_type=self.caldari_carrier_skill_type,
            required_level=3,
        )

        # when
        result = self.character.update_skill_sets()

        # then
        self.assertTrue(result.is_updated)

        self.assertEqual(self.character.skill_set_checks.count(), 1)
        first = self.character.skill_set_checks.first()
        self.assertEqual(first.skill_set.pk, skill_set.pk)
        skill_pks = {obj.pk for obj in first.failed_required_skills.all()}
        self.assertSetEqual(skill_pks, {skill_1.pk, skill_2.pk})

    def test_should_handle_no_skills(self):
        # when
        result = self.character.update_skill_sets()
        # then
        self.assertTrue(result.is_updated)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterStandingManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.character_1002 = create_memberaudit_character(1002)

    def test_can_create_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterStanding.objects.update_or_create_esi(self.character_1001)
        # then
        self.assertEqual(self.character_1001.standings.count(), 3)

        entry = self.character_1001.standings.get(eve_entity_id=1901)
        self.assertEqual(entry.standing, 0.1)

        entry = self.character_1001.standings.get(eve_entity_id=2901)
        self.assertEqual(entry.standing, 0)

        entry = self.character_1001.standings.get(eve_entity_id=500001)
        self.assertEqual(entry.standing, -1)

    def test_can_update_existing(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        npc_corp = EveEntity.objects.get(id=2901)
        create_character_standing(self.character_1001, npc_corp, standing=-5)
        # when
        CharacterStanding.objects.update_or_create_esi(self.character_1001)
        # then
        self.assertEqual(self.character_1001.standings.count(), 3)

        entry = self.character_1001.standings.get(eve_entity_id=1901)
        self.assertEqual(entry.standing, 0.1)

        entry = self.character_1001.standings.get(eve_entity_id=2901)
        self.assertEqual(entry.standing, 0)

        entry = self.character_1001.standings.get(eve_entity_id=500001)
        self.assertEqual(entry.standing, -1)

    def test_can_handle_no_standings(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterStanding.objects.update_or_create_esi(self.character_1002)
        # then
        self.assertEqual(self.character_1002.standings.count(), 0)

    def test_can_remove_obsolete_standings(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        obsolete_standing = create_character_standing(
            self.character_1001, EveEntity.objects.get(id=1101), standing=-5
        )
        # when
        CharacterStanding.objects.update_or_create_esi(self.character_1001)

        # then
        self.assertEqual(self.character_1001.standings.count(), 3)

        entry = self.character_1001.standings.get(eve_entity_id=1901)
        self.assertEqual(entry.standing, 0.1)

        entry = self.character_1001.standings.get(eve_entity_id=2901)
        self.assertEqual(entry.standing, 0)

        entry = self.character_1001.standings.get(eve_entity_id=500001)
        self.assertEqual(entry.standing, -1)

        self.assertFalse(
            self.character_1001.standings.filter(
                eve_entity_id=obsolete_standing.eve_entity.id
            ).exists()
        )


@patch(MANAGERS_PATH + ".esi")
class TestCharacterTitleManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    def test_should_add_new_title_from_scratch(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_titles",
                "character_id",
                needs_token=True,
                data={"1001": [{"name": "Awesome Title", "title_id": 1}]},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        self.character_1001.update_titles()
        # then
        self.assertEqual(self.character_1001.titles.count(), 1)
        obj = self.character_1001.titles.first()
        self.assertEqual(obj.name, "Awesome Title")
        self.assertEqual(obj.title_id, 1)

    def test_should_update_existing_titles(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_titles",
                "character_id",
                needs_token=True,
                data={"1001": [{"name": "Awesome Title", "title_id": 1}]},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        create_character_title(
            character=self.character_1001, name="Old title", title_id=1
        )
        # when
        self.character_1001.update_titles()
        # then
        obj = self.character_1001.titles.get(title_id=1)
        self.assertEqual(obj.name, "Awesome Title")

    def test_should_replace_existing_titles(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_titles",
                "character_id",
                needs_token=True,
                data={"1001": [{"name": "Awesome Title", "title_id": 2}]},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        create_character_title(
            character=self.character_1001, name="Old title", title_id=1
        )
        # when
        self.character_1001.update_titles()
        # then
        self.assertEqual(self.character_1001.titles.count(), 1)
        obj = self.character_1001.titles.get(title_id=2)
        self.assertEqual(obj.name, "Awesome Title")

    def test_should_remove_existing_titles(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_titles",
                "character_id",
                needs_token=True,
                data={"1001": []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        create_character_title(
            character=self.character_1001, name="Old title", title_id=1
        )
        # when
        self.character_1001.update_titles()
        # then
        self.assertEqual(self.character_1001.titles.count(), 0)

    def test_should_remove_xml_from_titles_and_strip(self, mock_esi):
        # given
        endpoints = [
            EsiEndpoint(
                "Character",
                "get_characters_character_id_titles",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {"name": "<color=0xFFee82ee> Awesome Title ", "title_id": 1}
                    ]
                },
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        self.character_1001.update_titles()
        # then
        self.assertEqual(self.character_1001.titles.count(), 1)
        obj = self.character_1001.titles.first()
        self.assertEqual(obj.name, "Awesome Title")
        self.assertEqual(obj.title_id, 1)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterWalletBalanceManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    def test_update_wallet_balance(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterWalletBalance.objects.update_or_create_esi(self.character_1001)
        # then
        self.assertEqual(self.character_1001.wallet_balance.total, 123456789)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MANAGERS_PATH + ".esi")
class TestCharacterWalletJournalManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    @patch(MANAGERS_PATH + ".data_retention_cutoff", lambda: None)
    def test_update_wallet_journal_1(self, mock_esi):
        """can create wallet journal entry from scratch"""
        mock_esi.client = esi_client_stub

        self.character_1001.update_wallet_journal()

        self.assertSetEqual(
            set(self.character_1001.wallet_journal.values_list("entry_id", flat=True)),
            {89, 91},
        )
        obj = self.character_1001.wallet_journal.get(entry_id=89)
        self.assertEqual(obj.amount, -100_000)
        self.assertEqual(float(obj.balance), 500_000.43)
        self.assertEqual(obj.context_id, 4)
        self.assertEqual(obj.context_id_type, obj.CONTEXT_ID_TYPE_CONTRACT_ID)
        self.assertEqual(obj.date, parse_datetime("2018-02-23T14:31:32Z"))
        self.assertEqual(obj.description, "Contract Deposit")
        self.assertEqual(obj.first_party.id, 2001)
        self.assertEqual(obj.reason, "just for fun")
        self.assertEqual(obj.ref_type, "contract_deposit")
        self.assertEqual(obj.second_party.id, 2002)

        obj = self.character_1001.wallet_journal.get(entry_id=91)
        self.assertEqual(
            obj.ref_type, "agent_mission_time_bonus_reward_corporation_tax"
        )

    @patch(MANAGERS_PATH + ".data_retention_cutoff", lambda: None)
    def test_update_wallet_journal_2(self, mock_esi):
        """can add entry to existing wallet journal"""
        mock_esi.client = esi_client_stub
        create_character_wallet_journal_entry(
            character=self.character_1001,
            entry_id=1,
            amount=1_000_000,
            balance=10_000_000,
            context_id_type=CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED,
            date=now(),
            description="dummy",
            first_party=EveEntity.objects.get(id=1001),
            second_party=EveEntity.objects.get(id=1002),
        )

        self.character_1001.update_wallet_journal()

        self.assertSetEqual(
            set(self.character_1001.wallet_journal.values_list("entry_id", flat=True)),
            {1, 89, 91},
        )

        obj = self.character_1001.wallet_journal.get(entry_id=89)
        self.assertEqual(obj.amount, -100_000)
        self.assertEqual(float(obj.balance), 500_000.43)
        self.assertEqual(obj.context_id, 4)
        self.assertEqual(obj.context_id_type, obj.CONTEXT_ID_TYPE_CONTRACT_ID)
        self.assertEqual(obj.date, parse_datetime("2018-02-23T14:31:32Z"))
        self.assertEqual(obj.description, "Contract Deposit")
        self.assertEqual(obj.first_party.id, 2001)
        self.assertEqual(obj.ref_type, "contract_deposit")
        self.assertEqual(obj.second_party.id, 2002)

    @patch(MANAGERS_PATH + ".data_retention_cutoff", lambda: None)
    def test_update_wallet_journal_3(self, mock_esi):
        """does not update existing entries"""
        mock_esi.client = esi_client_stub
        create_character_wallet_journal_entry(
            character=self.character_1001,
            entry_id=89,
            amount=1_000_000,
            balance=10_000_000,
            context_id_type=CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED,
            date=now(),
            description="dummy",
            first_party=EveEntity.objects.get(id=1001),
            second_party=EveEntity.objects.get(id=1002),
        )

        self.character_1001.update_wallet_journal()

        self.assertSetEqual(
            set(self.character_1001.wallet_journal.values_list("entry_id", flat=True)),
            {89, 91},
        )
        obj = self.character_1001.wallet_journal.get(entry_id=89)
        self.assertEqual(obj.amount, 1_000_000)
        self.assertEqual(float(obj.balance), 10_000_000)
        self.assertEqual(
            obj.context_id_type, CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED
        )
        self.assertEqual(obj.description, "dummy")
        self.assertEqual(obj.first_party.id, 1001)
        self.assertEqual(obj.second_party.id, 1002)

    def test_update_wallet_journal_4(self, mock_esi):
        """When new wallet entry is older than retention limit, then do not store it"""
        mock_esi.client = esi_client_stub

        with patch(
            MANAGERS_PATH + ".data_retention_cutoff",
            lambda: dt.datetime(2018, 3, 11, 20, 5, tzinfo=dt.timezone.utc)
            - dt.timedelta(days=10),
        ):
            self.character_1001.update_wallet_journal()

        self.assertSetEqual(
            set(self.character_1001.wallet_journal.values_list("entry_id", flat=True)),
            {91},
        )

    def test_update_wallet_journal_5(self, mock_esi):
        """When wallet existing entry is older than retention limit, then delete it"""
        mock_esi.client = esi_client_stub
        create_character_wallet_journal_entry(
            character=self.character_1001,
            entry_id=55,
            amount=1_000_000,
            balance=10_000_000,
            context_id_type=CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED,
            date=dt.datetime(2018, 2, 11, 20, 5, tzinfo=dt.timezone.utc),
            description="dummy",
            first_party=EveEntity.objects.get(id=1001),
            second_party=EveEntity.objects.get(id=1002),
        )

        with patch(
            MANAGERS_PATH + ".data_retention_cutoff",
            lambda: dt.datetime(2018, 3, 11, 20, 5, tzinfo=dt.timezone.utc)
            - dt.timedelta(days=20),
        ):
            self.character_1001.update_wallet_journal()

        self.assertSetEqual(
            set(self.character_1001.wallet_journal.values_list("entry_id", flat=True)),
            {89, 91},
        )


@patch(MANAGERS_PATH + ".esi")
class TestCharacterWalletTransactionManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character_1001 = create_memberaudit_character(1001)

    def test_should_add_wallet_transactions_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        with patch(MANAGERS_PATH + ".data_retention_cutoff", lambda: None):
            CharacterWalletTransaction.objects.update_or_create_esi(self.character_1001)
        # then
        expected = set(
            self.character_1001.wallet_transactions.values_list(
                "transaction_id", flat=True
            )
        )
        self.assertSetEqual(expected, {42})
        obj = self.character_1001.wallet_transactions.get(transaction_id=42)
        self.assertEqual(obj.client, EveEntity.objects.get(id=1003))
        self.assertEqual(obj.date, parse_datetime("2016-10-24T09:00:00Z"))
        self.assertTrue(obj.is_buy)
        self.assertTrue(obj.is_personal)
        self.assertIsNone(obj.journal_ref)
        self.assertEqual(obj.location, Location.objects.get(id=60003760))
        self.assertEqual(obj.quantity, 3)
        self.assertEqual(obj.eve_type, EveType.objects.get(id=603))
        self.assertEqual(float(obj.unit_price), 450000.99)

    def test_should_add_wallet_transactions_from_scratch_with_journal_ref(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        journal_entry = create_character_wallet_journal_entry(
            character=self.character_1001,
            entry_id=67890,
            amount=450000.99,
            balance=10_000_000,
            context_id_type=CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED,
            date=parse_datetime("2016-10-24T09:00:00Z"),
            description="dummy",
            first_party=EveEntity.objects.get(id=1001),
            second_party=EveEntity.objects.get(id=1003),
        )
        # when
        with patch(MANAGERS_PATH + ".data_retention_cutoff", lambda: None):
            CharacterWalletTransaction.objects.update_or_create_esi(self.character_1001)
        # then
        expected = set(
            self.character_1001.wallet_transactions.values_list(
                "transaction_id", flat=True
            )
        )

        self.assertSetEqual(expected, {42})
        obj = self.character_1001.wallet_transactions.get(transaction_id=42)
        self.assertEqual(obj.journal_ref, journal_entry)
